from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.compat import IS_PY2
from yafowil.file import convert_bytes
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.tests import wrapped_fxml

if IS_PY2:  # pragma: no cover
    from StringIO import StringIO
else:  # pragma: no cover
    from io import StringIO


class TestFile(YafowilTestCase):

    def test_file_blueprint(self):
        # Render file input
        widget = factory(
            'file',
            name='MYFILE')
        self.assertEqual(
            widget(),
            '<input id="input-MYFILE" name="MYFILE" type="file" />'
        )

        # Extract empty
        request = {}
        data = widget.extract(request)
        self.assertEqual(data.extracted, UNSET)

        # Extract ``new``
        request = {
            'MYFILE': {'file': StringIO('123')},
        }
        data = widget.extract(request)

        self.assertEqual(data.name, 'MYFILE')

        self.assertEqual(data.value, UNSET)

        self.assertEqual(sorted(data.extracted.keys()), ['action', 'file'])
        self.assertEqual(data.extracted['action'], 'new')
        self.assertTrue(isinstance(data.extracted['file'], StringIO))
        self.assertEqual(data.extracted['file'].read(), '123')

        self.assertEqual(data.errors, [])

        # File with value preset
        widget = factory(
            'file',
            name='MYFILE',
            value={
                'file': StringIO('321'),
            })
        self.checkOutput("""
        <div>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
          <div id="radio-MYFILE-keep">
            <input checked="checked" id="input-MYFILE-keep"
                   name="MYFILE-action" type="radio" value="keep"/>
            <span>Keep Existing file</span>
          </div>
          <div id="radio-MYFILE-replace">
            <input id="input-MYFILE-replace" name="MYFILE-action"
                   type="radio" value="replace"/>
            <span>Replace existing file</span>
          </div>
          <div id="radio-MYFILE-delete">
            <input id="input-MYFILE-delete" name="MYFILE-action"
                   type="radio" value="delete"/>
            <span>Delete existing file</span>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Extract ``keep`` returns original value
        request = {
            'MYFILE': {'file': StringIO('123')},
            'MYFILE-action': 'keep'
        }
        data = widget.extract(request)

        self.assertEqual(data.name, 'MYFILE')

        self.assertEqual(sorted(data.value.keys()), ['action', 'file'])
        self.assertEqual(data.value['action'], 'keep')
        self.assertTrue(isinstance(data.value['file'], StringIO))

        self.assertEqual(sorted(data.extracted.keys()), ['action', 'file'])
        self.assertEqual(data.extracted['action'], 'keep')
        self.assertTrue(isinstance(data.extracted['file'], StringIO))
        self.assertEqual(data.extracted['file'].read(), '321')

        self.assertEqual(data.errors, [])

        # Extract ``replace`` returns new value
        request['MYFILE-action'] = 'replace'
        data = widget.extract(request)

        self.assertEqual(sorted(data.extracted.keys()), ['action', 'file'])
        self.assertEqual(data.extracted['action'], 'replace')
        self.assertEqual(data.extracted['file'].read(), '123')

        # Extract empty ``replace`` results in extraction error
        request = {
            'MYFILE': '',
            'MYFILE-action': 'replace'
        }
        data = widget.extract(request)

        self.assertEqual(data.errors, [
            ExtractionError('Cannot replace file. No file uploaded.')
        ])
        self.assertEqual(data.extracted, UNSET)

        # Extract ``delete`` returns UNSET
        request['MYFILE-action'] = 'delete'
        data = widget.extract(request)
        self.assertEqual(
            data.extracted,
            {'action': 'delete', 'file': UNSET}
        )

        self.assertEqual(data.extracted['action'], 'delete')

        self.checkOutput("""
        <div>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
          <div id="radio-MYFILE-keep">
            <input id="input-MYFILE-keep" name="MYFILE-action" type="radio"
                   value="keep"/>
            <span>Keep Existing file</span>
          </div>
          <div id="radio-MYFILE-replace">
            <input id="input-MYFILE-replace" name="MYFILE-action" type="radio"
                   value="replace"/>
            <span>Replace existing file</span>
          </div>
          <div id="radio-MYFILE-delete">
            <input checked="checked" id="input-MYFILE-delete"
                   name="MYFILE-action" type="radio" value="delete"/>
            <span>Delete existing file</span>
          </div>
        </div>
        """, wrapped_fxml(widget(request=request)))

        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'foo/bar'
            })
        self.assertEqual(widget(), (
            '<input accept="foo/bar" id="input-MYFILE" '
            'name="MYFILE" type="file" />'
        ))

        # File actions vocabulary
        widget = factory(
            'file',
            name='MYFILE',
            value={
                'file': StringIO('321')
            },
            props={
                'vocabulary': [
                    ('keep', 'Keep Existing file'),
                    ('replace', 'Replace existing file')
                ]
            })
        self.checkOutput("""
        <div>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
          <div id="radio-MYFILE-keep">
            <input checked="checked" id="input-MYFILE-keep"
                   name="MYFILE-action" type="radio" value="keep"/>
            <span>Keep Existing file</span>
          </div>
          <div id="radio-MYFILE-replace">
            <input id="input-MYFILE-replace" name="MYFILE-action"
                   type="radio" value="replace"/>
            <span>Replace existing file</span>
          </div>
        </div>
        """, wrapped_fxml(widget()))

        # Mimetype extractor
        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': '*/*'
            })
        request = {
            'MYFILE': {
                'file': StringIO('123'),
                'mimetype': 'image/jpeg'
            }
        }
        data = widget.extract(request)
        expected = {
            'action': 'new',
            'file': request['MYFILE']['file'],
            'mimetype': 'image/jpeg'
        }
        self.assertEqual(data.extracted, expected)

        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'image/*'
            })
        data = widget.extract(request)
        self.assertEqual(data.extracted, expected)

        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'image/png,image/jpeg'
            })
        data = widget.extract(request)
        self.assertEqual(data.extracted, expected)

        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'video/webm,image/*'
            })
        data = widget.extract(request)
        self.assertEqual(data.extracted, expected)

        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'image/png'
            })
        data = widget.extract(request)
        self.assertEqual(data.errors, [
            ExtractionError('Mimetype of uploaded file not matches')
        ])

        # no validation if mimetype not on extracted file
        # happens with keep and replace actions.
        widget = factory(
            'file',
            name='MYFILE',
            value=dict(),
            props={
                'accept': 'image/png'
            })
        request = {
            'MYFILE': {},
            'MYFILE-action': 'keep'
        }
        data = widget.extract(request)
        self.assertEqual(data.extracted, {'action': 'keep'})

        request['MYFILE-action'] = 'delete'
        data = widget.extract(request)
        self.assertEqual(data.extracted, {'action': 'delete', 'file': UNSET})

        # File display renderer
        self.assertEqual(convert_bytes(1 * 1024 * 1024 * 1024 * 1024), '1.00T')
        self.assertEqual(convert_bytes(1 * 1024 * 1024 * 1024), '1.00G')
        self.assertEqual(convert_bytes(1 * 1024 * 1024), '1.00M')
        self.assertEqual(convert_bytes(1 * 1024), '1.00K')
        self.assertEqual(convert_bytes(1), '1.00b')

        widget = factory(
            'file',
            name='MYFILE',
            mode='display')
        self.checkOutput("""
        <div>No file</div>
        """, fxml(widget()))

        value = {
            'file': StringIO('12345'),
            'mimetype': 'text/plain',
            'filename': 'foo.txt',
        }
        widget = factory(
            'file',
            name='MYFILE',
            value=value,
            mode='display')
        self.checkOutput("""
        <div>
          <ul>
            <li><strong>Filename: </strong>foo.txt</li>
            <li><strong>Mimetype: </strong>text/plain</li>
            <li><strong>Size: </strong>5.00b</li>
          </ul>
        </div>
        """, fxml(widget()))

        # Generic HTML5 Data
        widget = factory(
            'file',
            name='MYFILE',
            props={
                'accept': 'foo/bar',
                'data': {
                    'foo': 'bar'
                }
            })
        self.assertEqual(widget(), (
            '<input accept="foo/bar" data-foo=\'bar\' id="input-MYFILE" '
            'name="MYFILE" type="file" />'
        ))

        widget.mode = 'display'
        self.assertEqual(widget(), "<div data-foo='bar'>No file</div>")



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestPassword(YafowilTestCase):

    def test_password_blueprint(self):
        # Password widget has some additional properties, ``strength``,
        # ``minlength`` and ``ascii``.

        # Use in add forms, no password set yet
        widget = factory(
            'password',
            name='PWD')
        self.assertEqual(widget(), (
            '<input class="password" id="input-PWD" name="PWD" '
            'type="password" value="" />'
        ))

        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'PWD': 'xx'})
        self.assertEqual(data.extracted, 'xx')

        widget.mode = 'display'
        self.assertEqual(widget(), '')

        # Use in edit forms. note that password is never shown up in markup,
        # but a placeholder is used when a password is already set. Thus, if a
        # extracted password value is UNSET, this means that password was not
        # changed
        widget = factory(
            'password',
            name='PASSWORD',
            value='secret')
        self.assertEqual(widget(), (
            '<input class="password" id="input-PASSWORD" name="PASSWORD" '
            'type="password" value="_NOCHANGE_" />'
        ))

        data = widget.extract({'PASSWORD': '_NOCHANGE_'})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'PASSWORD': 'foo'})
        self.assertEqual(data.extracted, 'foo')

        self.assertEqual(widget(data=data), (
            '<input class="password" id="input-PASSWORD" name="PASSWORD" '
            'type="password" value="foo" />'
        ))

        widget.mode = 'display'
        self.assertEqual(widget(), '********')

        # Password validation
        widget = factory(
            'password',
            name='PWD',
            props={
                'strength': 5,  # max 4, does not matter, max is used
            })
        data = widget.extract({'PWD': ''})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': 'A0*'})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': 'a0*'})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': 'aA*'})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': 'aA0'})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': 'aA0*'})
        self.assertEqual(data.errors, [])

        # Minlength validation
        widget = factory(
            'password',
            name='PWD',
            props={
                'minlength': 3,
            })
        data = widget.extract({'PWD': 'xx'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input must have at least 3 characters.')]
        )

        data = widget.extract({'PWD': 'xxx'})
        self.assertEqual(data.errors, [])

        # Ascii validation
        widget = factory(
            'password',
            name='PWD',
            props={
                'ascii': True,
            })
        data = widget.extract({'PWD': u'äää'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input contains illegal characters.')]
        )

        data = widget.extract({'PWD': u'xx'})
        self.assertEqual(data.errors, [])

        # Combine all validations
        widget = factory(
            'password',
            name='PWD',
            props={
                'required': 'No Password given',
                'minlength': 6,
                'ascii': True,
                'strength': 4,
            })
        data = widget.extract({'PWD': u''})
        self.assertEqual(data.errors, [ExtractionError('No Password given')])

        data = widget.extract({'PWD': u'xxxxx'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input must have at least 6 characters.')]
        )

        data = widget.extract({'PWD': u'xxxxxä'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input contains illegal characters.')]
        )

        data = widget.extract({'PWD': u'xxxxxx'})
        self.assertEqual(data.errors, [ExtractionError('Password too weak')])

        data = widget.extract({'PWD': u'xX1*00'})
        self.assertEqual(data.errors, [])

        # Emptyvalue
        widget = factory(
            'password',
            name='PWD',
            props={
                'emptyvalue': 'DEFAULTPWD',  # <- not a good idea, but works
            })
        data = widget.extract(request={'PWD': ''})
        self.assertEqual(data.name, 'PWD')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'DEFAULTPWD')
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'PWD': 'NOEMPTY'})
        self.assertEqual(data.name, 'PWD')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'NOEMPTY')
        self.assertEqual(data.errors, [])

        # Persist
        widget = factory(
            'password',
            name='PWD',
            props={
                'persist_writer': write_mapping_writer
            })
        data = widget.extract(request={'PWD': '1234'})
        model = dict()
        data.write(model)
        self.assertEqual(model, {'PWD': '1234'})



from node.utils import UNSET
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.tests import wrapped_fxml


class TestLines(YafowilTestCase):

    def test_lines_blueprint(self):
        # Render empty
        widget = factory(
            'lines',
            name='MYLINES',
            value=None)
        self.assertEqual(widget(), (
            '<textarea class="lines" cols="40" id="input-MYLINES" '
            'name="MYLINES" rows="8"></textarea>'
        ))

        # Render with preset value, expected as list
        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'])
        self.checkOutput("""
        <textarea class="lines" cols="40" id="input-MYLINES"
                  name="MYLINES" rows="8">a
        b
        c</textarea>
        """, fxml(widget()))

        # Extract empty
        data = widget.extract({'MYLINES': ''})
        self.assertEqual(data.extracted, [])

        # Extract with data
        data = widget.extract({'MYLINES': 'a\nb'})
        self.assertEqual(data.extracted, ['a', 'b'])

        # Render with extracted data
        self.checkOutput("""
        <textarea class="lines" cols="40" id="input-MYLINES"
                  name="MYLINES" rows="8">a
        b</textarea>
        """, fxml(widget(data=data)))

        # Display mode with preset value
        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'],
            mode='display')
        self.checkOutput("""
        <ul class="display-lines" id="display-MYLINES">
          <li>a</li>
          <li>b</li>
          <li>c</li>
        </ul>
        """, fxml(widget()))

        # Display mode with empty preset value
        widget = factory(
            'lines',
            name='MYLINES',
            value=[],
            mode='display')
        self.checkOutput("""
        <ul class="display-lines" id="display-MYLINES"/>
        """, fxml(widget()))

        # Display mode with ``display_proxy``
        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'],
            props={
                'display_proxy': True,
            },
            mode='display')
        self.checkOutput("""
        <div>
          <ul class="display-lines" id="display-MYLINES">
            <li>a</li>
            <li>b</li>
            <li>c</li>
          </ul>
          <input class="lines" id="input-MYLINES" name="MYLINES" type="hidden"
                 value="a"/>
          <input class="lines" id="input-MYLINES" name="MYLINES" type="hidden"
                 value="b"/>
          <input class="lines" id="input-MYLINES" name="MYLINES" type="hidden"
                 value="c"/>
        </div>
        """, wrapped_fxml(widget()))

        data = widget.extract({'MYLINES': 'a\nb'})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, ['a', 'b', 'c'])
        self.assertEqual(data.extracted, ['a', 'b'])
        self.assertEqual(data.errors, [])

        self.checkOutput("""
        <div>
          <ul class="display-lines" id="display-MYLINES">
            <li>a</li>
            <li>b</li>
          </ul>
          <input class="lines" id="input-MYLINES" name="MYLINES" type="hidden"
                 value="a"/>
          <input class="lines" id="input-MYLINES" name="MYLINES" type="hidden"
                 value="b"/>
        </div>
        """, wrapped_fxml(widget(data=data)))

        # Generic HTML5 Data
        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'],
            props={
                'data': {'foo': 'bar'}
            })
        self.checkOutput("""
        <textarea class="lines" cols="40" data-foo="bar" id="input-MYLINES"
                  name="MYLINES" rows="8">a
        b
        c</textarea>
        """, fxml(widget()))

        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'],
            props={
                'data': {'foo': 'bar'}
            },
            mode='display')
        self.checkOutput("""
        <ul class="display-lines" data-foo="bar" id="display-MYLINES">
          <li>a</li>
          <li>b</li>
          <li>c</li>
        </ul>
        """, fxml(widget()))

        # Emptyvalue
        widget = factory(
            'lines',
            name='MYLINES',
            value=['a', 'b', 'c'],
            props={
                'emptyvalue': ['1']
            })
        data = widget.extract(request={'MYLINES': ''})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, ['a', 'b', 'c'])
        self.assertEqual(data.extracted, ['1'])
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'MYLINES': '1\n2'})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, ['a', 'b', 'c'])
        self.assertEqual(data.extracted, ['1', '2'])
        self.assertEqual(data.errors, [])

        # Datatype
        widget = factory(
            'lines',
            name='MYLINES',
            props={
                'emptyvalue': [1],
                'datatype': int
            })
        data = widget.extract(request={'MYLINES': ''})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [1])
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'MYLINES': '1\n2'})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [1, 2])
        self.assertEqual(data.errors, [])

        widget.attrs['emptyvalue'] = ['1']
        data = widget.extract(request={'MYLINES': ''})
        self.assertEqual(data.name, 'MYLINES')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, [1])
        self.assertEqual(data.errors, [])

        # Persist
        widget = factory(
            'lines',
            name='MYLINES',
            props={
                'persist_writer': write_mapping_writer
            })
        data = widget.extract(request={'MYLINES': '1\n2'})
        model = dict()
        data.write(model)
        self.assertEqual(model, {'MYLINES': ['1', '2']})



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.common import empty_display_renderer
from yafowil.common import input_attributes_common
from yafowil.tsf import _
from yafowil.utils import attr_value
from yafowil.utils import managedprops


###############################################################################
# submit
###############################################################################

@managedprops(
    'tag_type',
    'label',
    'class',
    'action',
    'handler',
    'next',
    'skip',
    'expression')
def submit_renderer(widget, data):
    expression = attr_value('expression', widget, data)
    if not expression:
        return u''
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'submit'
    input_attrs['name_'] = attr_value('action', widget, data) \
        and 'action.{0}'.format(widget.dottedpath)
    input_attrs['value'] = attr_value('label', widget, data, widget.name)
    return tag(attr_value('tag_type', widget, data), **input_attrs)


factory.register(
    'submit',
    edit_renderers=[submit_renderer],
    display_renderers=[empty_display_renderer]
)

factory.doc['blueprint']['submit'] = """\
Submit action inside the form
"""

factory.doc['props']['submit.label'] = """\
Label of the submit.
"""

factory.defaults['submit.tag_type'] = 'input'
factory.doc['props']['submit.tag_type'] = """\
Define the type of tag that will be rendered.
"""

factory.defaults['submit.expression'] = True
factory.doc['props']['submit.expression'] = """\
Flag or expression callable whether this action is available to the user
or not.
"""

factory.defaults['submit.action'] = True
factory.doc['props']['submit.action'] = """\
Marks this widget as an action. One out of ``True`` or ``False``.
"""

factory.defaults['submit.skip'] = False
factory.doc['props']['submit.skip'] = """\
Skips action and only perform next. One out of ``True`` or ``False``.
"""

factory.doc['props']['submit.handler'] = """\
Handler is a callable which get called if this action performs. It expects two
parameters: ``widget``, ``data``.
"""

factory.doc['props']['submit.next'] = """\
Next is a callable expected to return the web address. It expects a request as
the only parameter.
"""

factory.defaults['submit.disabled'] = False
factory.doc['props']['submit.disabled'] = """\
Flag the submit field as disabled.
"""


###############################################################################
# button
###############################################################################

@managedprops(
    'text',
    'action',
    'handler',
    'next',
    'skip',
    'type',
    'expression',
    'form',
    'formaction',
    'formenctype',
    'formmethod',
    'formnovalidate',
    'formtarget',
    'autofocus',
    'disabled',
    'class',
    'class_add',
    'accesskey')
def button_renderer(widget, data):
    expression = attr_value('expression', widget, data)
    if not expression:
        return u''
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = attr_value('type', widget, data)
    input_attrs['form'] = attr_value('form', widget, data)
    input_attrs['formaction'] = attr_value('formaction', widget, data)
    input_attrs['formenctype'] = attr_value('formenctype', widget, data)
    input_attrs['formmethod'] = attr_value('formmethod', widget, data)
    input_attrs['formnovalidate'] = attr_value('formnovalidate', widget, data)
    input_attrs['formtarget'] = attr_value('formtarget', widget, data)
    input_attrs['accesskey'] = attr_value('accesskey', widget, data)
    if input_attrs['type'] == 'submit':
        # only 'submit' type sends data to the server
        if attr_value('action', widget, data):
            # only if an server side action is set this makes sense:
            input_attrs['name_'] = 'action.{0}'.format(widget.dottedpath)
        else:
            # otherwise check for a custom name
            input_attrs['name_'] = attr_value('name', widget, data)
    text = attr_value('text', widget, data)
    return tag("button", text, **input_attrs)


factory.register(
    'button',
    edit_renderers=[button_renderer],
    display_renderers=[empty_display_renderer]
)

factory.defaults['button.type'] = 'submit'

factory.doc['blueprint']['button'] = """\
Represents a clickable button, used to submit forms or anywhere in a document for
accessible, standard button functionality.
"""

factory.doc['props']['button.text'] = """\
The content of the button element.
"""

# parts of the docs are copied over from
# https://developer.mozilla.org/de/docs/Web/HTML/Element/button

factory.doc['props']['button.form'] = """\
The <form> element to associate the button with (its form owner).
The value of this attribute must be the id of a <form> in the same document.
(If this attribute is not set, the <button> is associated with its ancestor
<form> element, if any.)
"""

factory.doc['props']['button.formaction'] = """\
The URL that processes the information submitted by the button.
Overrides the action attribute of the button's form owner.
Does nothing if there is no form owner.
"""

factory.doc['props']['button.formenctype'] = """\
If the button is a submit button (it's inside/associated with a <form> and
doesn't have type="button"), specifies how to encode the form data that is
submitted. Possible values:

application/x-www-form-urlencoded: The default if the attribute is not used.

multipart/form-data: Use to submit <input> elements with their type attributes
set to file.

text/plain: Specified as a debugging aid; shouldn’t be used for real form
submission.

If this attribute is specified, it overrides the enctype attribute of the
button's form owner.
"""

factory.doc['props']['button.formmethod'] = """\
If the button is a submit button (it's inside/associated with a <form> and
doesn't have type="button"), this attribute specifies the HTTP method used
to submit the form. Possible values: POST or GET. If specified, this
attribute overrides the method attribute of the button's form owner.
"""

factory.doc['props']['button.formnovalidate'] = """\
If the button is a submit button, this Boolean attribute specifies that the
form is not to be validated when it is submitted. If this attribute is
specified, it overrides the novalidate attribute of the button's form owner.
"""

factory.doc['props']['button.formtarget'] = """\
If the button is a submit button, this attribute is a author-defined name or
standardized, underscore-prefixed keyword indicating where to display the
response from submitting the form. This is the name of, or keyword for,
a browsing context (a tab, window, or <iframe>). If this attribute is
specified, it overrides the target attribute of the button's form owner.
The following keywords have special meanings:

_self: Load the response into the same browsing context as the current one.
This is the default if the attribute is not specified.

_blank: Load the response into a new unnamed browsing context — usually a new
tab or window, depending on the user’s browser settings.

_parent: Load the response into the parent browsing context of the current one.
If there is no parent, this option behaves the same way as _self.

_top: Load the response into the top-level browsing context (that is, the
browsing context that is an ancestor of the current one, and has no parent).
If there is no parent, this option behaves the same way as _self.
"""

factory.defaults['button.expression'] = True
factory.doc['props']['button.expression'] = """\
Flag or expression callable whether this action is available to the user
or not.
"""

factory.defaults['button.action'] = True
factory.doc['props']['button.action'] = """\
Marks this widget as an action. One out of ``True`` or ``False``.
"""

factory.defaults['button.skip'] = False
factory.doc['props']['button.skip'] = """\
Skips action and only perform next. One out of ``True`` or ``False``.
"""

factory.doc['props']['button.handler'] = """\
Handler is a callable which get called if this action performs. It expects two
parameters: ``widget``, ``data``.
"""

factory.doc['props']['button.next'] = """\
Next is a callable expected to return the web address. It expects a request as
the only parameter.
"""

factory.defaults['button.disabled'] = False
factory.doc['props']['button.disabled'] = """\
Flag the button field as disabled.
"""

factory.defaults['button.autofocus'] = False
factory.doc['props']['button.autofocus'] = """\
Boolean attribute specifies that the button should have input focus when the
page loads. Only one element in a document can have this attribute.
"""

factory.defaults['button.autocomplete'] = False
factory.doc['props']['button.autocomplete'] = """\
This attribute on a <button> is nonstandard and Firefox-specific.
Unlike other browsers, Firefox persists the dynamic disabled state of a
<button> across page loads. Setting autocomplete="off" on the button
disables this feature
"""



from node.utils import UNSET
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
from yafowil.tests import wrapped_fxml


class TestCheckbox(YafowilTestCase):

    def test_checkbox_blueprint(self):
        # A boolean checkbox widget (default)
        widget = factory(
            'checkbox',
            name='MYCHECKBOX')
        self.checkOutput("""
        <div>
          <input class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
                 type="checkbox" value=""/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-checkbox" id="display-MYCHECKBOX">No</div>'
        )

        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='True')
        self.checkOutput("""
        <div>
          <input checked="checked" class="checkbox" id="input-MYCHECKBOX"
                 name="MYCHECKBOX" type="checkbox" value=""/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-checkbox" id="display-MYCHECKBOX">Yes</div>'
        )

        # A checkbox with custom CSS related empty label
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            props={
                'with_label': True
            })
        self.checkOutput("""
        <input class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
        type="checkbox" value="" /><label class="checkbox_label"
        for="input-MYCHECKBOX">&nbsp;</label><input
        id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists" type="hidden"
        value="checkboxexists" />
        """, widget())

        # A checkbox widget with a value or an empty string
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='',
            props={
                'format': 'string'
            })
        self.checkOutput("""
        <div>
          <input class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
                 type="checkbox" value=""/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.assertEqual(
            widget(),
            '<div class="display-checkbox" id="display-MYCHECKBOX">No</div>'
        )

        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='Test Checkbox',
            props={
                'format': 'string'
            })
        self.checkOutput("""
        <div>
          <input checked="checked" class="checkbox" id="input-MYCHECKBOX"
                 name="MYCHECKBOX" type="checkbox" value="Test Checkbox"/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        widget.mode = 'display'
        self.assertEqual(widget(), (
            '<div class="display-checkbox" id="display-MYCHECKBOX">'
            'Test Checkbox</div>'
        ))

        # Checkbox with manually set 'checked' attribute
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='',
            props={
                'format': 'string',
                'checked': True,
            })
        self.checkOutput("""
        <div>
          <input checked="checked" class="checkbox" id="input-MYCHECKBOX"
                 name="MYCHECKBOX" type="checkbox" value=""/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='Test Checkbox',
            props={
                'format': 'string',
                'checked': False,
            })
        self.checkOutput("""
        <div>
          <input class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
                 type="checkbox" value="Test Checkbox"/>
          <input id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists"
                 type="hidden" value="checkboxexists"/>
        </div>
        """, wrapped_fxml(widget()))

        # Checkbox extraction
        request = {
            'MYCHECKBOX': '1',
            'MYCHECKBOX-exists': 'checkboxexists'
        }
        data = widget.extract(request)
        self.checkOutput("""
        <RuntimeData MYCHECKBOX, value='Test Checkbox', extracted='1' at ...>
        """, data.treerepr())

        request = {
            'MYCHECKBOX': '',
            'MYCHECKBOX-exists': 'checkboxexists'
        }
        data = widget.extract(request)
        self.checkOutput("""
        <RuntimeData MYCHECKBOX, value='Test Checkbox', extracted='' at ...>
        """, data.treerepr())

        request = {
            'MYCHECKBOX': 1,
        }
        data = widget.extract(request)
        self.checkOutput("""
        <RuntimeData MYCHECKBOX, value='Test Checkbox', extracted=<UNSET> at ...>
        """, data.treerepr())  # noqa

        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'MYCHECKBOX': UNSET})

        # bool extraction
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='Test Checkbox',
            props={
                'format': 'bool'
            })
        request = {
            'MYCHECKBOX': '',
            'MYCHECKBOX-exists': 'checkboxexists'
        }
        data = widget.extract(request)
        self.checkOutput("""
        <RuntimeData MYCHECKBOX, value='Test Checkbox', extracted=True at ...>
        """, data.treerepr())

        request = {
            'MYCHECKBOX-exists': 'checkboxexists'
        }
        data = widget.extract(request)
        self.checkOutput("""
        <RuntimeData MYCHECKBOX, value='Test Checkbox', extracted=False at ...>
        """, data.treerepr())

        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'MYCHECKBOX': False})

        # invalid format
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            props={
                'format': 'invalid'
            })
        request = {
            'MYCHECKBOX': '',
            'MYCHECKBOX-exists': 'checkboxexists'
        }
        with self.assertRaises(ValueError) as arc:
            widget.extract(request)
        msg = "Checkbox widget has invalid format 'invalid' set"
        self.assertEqual(str(arc.exception), msg)

        # Render in display mode
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value=False,
            mode='display',
            props={
                'format': 'bool'
            })
        self.checkOutput("""
        <div>
          <div class="display-checkbox" id="display-MYCHECKBOX">No</div>
        </div>
        """, wrapped_fxml(widget()))

        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value=True,
            mode='display',
            props={
                'format': 'bool'
            })
        self.checkOutput("""
        <div>
          <div class="display-checkbox" id="display-MYCHECKBOX">Yes</div>
        </div>
        """, wrapped_fxml(widget()))

        # Display mode and display proxy bool format
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value=True,
            props={
                'format': 'bool',
                'display_proxy': True
            },
            mode='display')
        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">Yes<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX" type="hidden"
        value="" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget())

        data = widget.extract(request={'MYCHECKBOX-exists': 'checkboxexists'})
        self.assertEqual(data.name, 'MYCHECKBOX')
        self.assertEqual(data.value, True)
        self.assertEqual(data.extracted, False)

        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">No<input
        id="checkboxexists-MYCHECKBOX" name="MYCHECKBOX-exists" type="hidden"
        value="checkboxexists" /></div>
        """, widget(data=data))

        data = widget.extract(request={
            'MYCHECKBOX-exists': 'checkboxexists',
            'MYCHECKBOX': ''
        })
        self.assertEqual(data.name, 'MYCHECKBOX')
        self.assertEqual(data.value, True)
        self.assertEqual(data.extracted, True)

        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">Yes<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
        type="hidden" value="" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget(data=data))

        # Display mode and display proxy string format
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='yes',
            props={
                'format': 'string',
                'display_proxy': True
            },
            mode='display')
        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">yes<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
        type="hidden" value="yes" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget())

        data = widget.extract(request={'MYCHECKBOX-exists': 'checkboxexists'})
        self.assertEqual(data.name, 'MYCHECKBOX')
        self.assertEqual(data.value, 'yes')
        self.assertEqual(data.extracted, '')

        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">No<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX" type="hidden"
        value="" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget(data=data))

        data = widget.extract(request={
            'MYCHECKBOX-exists': 'checkboxexists',
            'MYCHECKBOX': ''
        })
        self.assertEqual(data.name, 'MYCHECKBOX')
        self.assertEqual(data.value, 'yes')
        self.assertEqual(data.extracted, '')

        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">No<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX" type="hidden"
        value="" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget(data=data))

        data = widget.extract(request={'MYCHECKBOX-exists': 'checkboxexists',
                                       'MYCHECKBOX': 'foo'})
        self.assertEqual(data.name, 'MYCHECKBOX')
        self.assertEqual(data.value, 'yes')
        self.assertEqual(data.extracted, 'foo')

        self.checkOutput("""
        <div class="display-checkbox" id="display-MYCHECKBOX">foo<input
        class="checkbox" id="input-MYCHECKBOX" name="MYCHECKBOX"
        type="hidden" value="foo" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" /></div>
        """, widget(data=data))

        # Generic HTML5 Data
        widget = factory(
            'checkbox',
            name='MYCHECKBOX',
            value='Test Checkbox',
            props={
                'data': {'foo': 'bar'}
            })
        self.checkOutput("""
        <input checked="checked" class="checkbox" data-foo=\'bar\'
        id="input-MYCHECKBOX" name="MYCHECKBOX" type="checkbox"
        value="" /><input id="checkboxexists-MYCHECKBOX"
        name="MYCHECKBOX-exists" type="hidden" value="checkboxexists" />
        """, widget())



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import generic_extractor
from yafowil.compat import BYTES_TYPE
from yafowil.compat import IS_PY2
from yafowil.compat import LONG_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.datatypes import convert_value_to_datatype
from yafowil.datatypes import convert_values_to_datatype
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.datatypes import lookup_datatype_converter
from yafowil.tests import YafowilTestCase
from yafowil.utils import EMPTY_VALUE
import copy
import uuid


class TestDatatypes(YafowilTestCase):

    def test_EMPTY_VALUE(self):
        self.assertEqual(repr(EMPTY_VALUE), '<EMPTY_VALUE>')
        self.assertEqual(str(EMPTY_VALUE), '')
        self.assertFalse(bool(EMPTY_VALUE))
        self.assertEqual(len(EMPTY_VALUE), 0)
        self.assertTrue(copy.copy(EMPTY_VALUE) is EMPTY_VALUE)
        self.assertTrue(copy.deepcopy(EMPTY_VALUE) is EMPTY_VALUE)
        self.assertFalse(EMPTY_VALUE < EMPTY_VALUE)
        self.assertFalse(EMPTY_VALUE <= EMPTY_VALUE)
        self.assertFalse(EMPTY_VALUE > EMPTY_VALUE)
        self.assertFalse(EMPTY_VALUE >= EMPTY_VALUE)


    def test_generic_emptyvalue_extractor(self):
        factory.register(
            'emptyvalue_test',
            extractors=[
                generic_extractor,
                generic_emptyvalue_extractor
            ]
        )
        widget = factory('emptyvalue_test', name='widget')
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'widget': ''})
        self.assertEqual(data.extracted, '')

        widget = factory(
            'emptyvalue_test',
            name='widget',
            props={'emptyvalue': EMPTY_VALUE}
        )
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'widget': ''})
        self.assertEqual(data.extracted, EMPTY_VALUE)

        widget = factory(
            'emptyvalue_test',
            name='widget',
            props={'emptyvalue': 'EMPTY'}
        )
        data = widget.extract({})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'widget': ''})
        self.assertEqual(data.extracted, 'EMPTY')

    def test_DatatypeConverter(self):
        from yafowil.datatypes import DatatypeConverter

        converter = DatatypeConverter()
        self.assertEqual(converter.to_form(u'"quoted"'), u'&quot;quoted&quot;')

        converter = DatatypeConverter(int)
        self.assertEqual(converter.to_form(1), '1')
        self.assertEqual(converter.to_value('1'), 1)
        self.assertEqual(converter.to_value(1), 1)
        self.assertEqual(converter.to_value(1.0), 1)

        converter = DatatypeConverter(uuid.UUID)
        uuid_ = uuid.uuid4()
        self.assertEqual(converter.to_form(uuid_), UNICODE_TYPE(uuid_))
        self.assertEqual(converter.to_value(UNICODE_TYPE(uuid_)), uuid_)
        self.assertEqual(converter.to_value(uuid_), uuid_)

    def test_BytesDatatypeConverter(self):
        from yafowil.datatypes import BytesDatatypeConverter

        converter = BytesDatatypeConverter()
        self.assertEqual(
            converter.to_form(b'\r\n\x01\x9a\x03\xff'),
            u'\\r\\n\\x01\\x9a\\x03\\xff'
        )
        self.assertEqual(
            converter.to_form(u'aaa'),
            u'aaa'
        )
        with self.assertRaises(UnicodeEncodeError):
            converter.to_value(u'ä')
        self.assertEqual(
            converter.to_value(u'\\r\\n\\x01\\x9a\\x03\\xff'),
            b'\r\n\x01\x9a\x03\xff'
        )
        self.assertEqual(
            converter.to_value(b'aaa'),
            b'aaa'
        )

    def test_UnicodeDatatypeConverter(self):
        from yafowil.datatypes import UnicodeDatatypeConverter

        converter = UnicodeDatatypeConverter()
        self.assertEqual(converter.to_value(u'äöü'), u'äöü')
        self.assertEqual(converter.to_value(b'aou'), u'aou')

    def test_FloatDatatypeConverter(self):
        from yafowil.datatypes import FloatDatatypeConverter

        converter = FloatDatatypeConverter()
        self.assertEqual(converter.to_value('1.1'), 1.1)
        self.assertEqual(converter.to_value('1,1'), 1.1)
        self.assertEqual(converter.to_value(1.1), 1.1)

    def test_lookup_datatype_converter(self):
        from yafowil.datatypes import DatatypeConverter
        converter = lookup_datatype_converter(int)
        self.assertIsInstance(converter, DatatypeConverter)
        self.assertEqual(converter.type_, int)

        from yafowil.datatypes import BytesDatatypeConverter

        converter = lookup_datatype_converter(BYTES_TYPE)
        self.assertIsInstance(converter, BytesDatatypeConverter)

        from yafowil.datatypes import UnicodeDatatypeConverter

        converter = lookup_datatype_converter(UNICODE_TYPE)
        self.assertIsInstance(converter, UnicodeDatatypeConverter)

        from yafowil.datatypes import FloatDatatypeConverter

        converter = lookup_datatype_converter(float)
        self.assertIsInstance(converter, FloatDatatypeConverter)

    def test_convert_value_to_datatype(self):
        # Unknown string identifier
        with self.assertRaises(KeyError) as arc:
            convert_value_to_datatype('val', 'inexistent')
        self.assertEqual(str(arc.exception), "'inexistent'")

        # Function returns ``EMPTY_VALUE`` marker if value is ``None`` or empty
        # string
        self.assertEqual(convert_value_to_datatype('', 'uuid'), EMPTY_VALUE)
        self.assertEqual(convert_value_to_datatype(None, 'uuid'), EMPTY_VALUE)

    def test_convert_value_to_datatype_bytes(self):
        # Convert to string by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'str'), UNSET)

        converted = convert_value_to_datatype(u'string', 'str')
        self.assertEqual(converted, b'string')
        self.assertTrue(isinstance(converted, BYTES_TYPE))

        with self.assertRaises(UnicodeEncodeError) as arc:
            convert_value_to_datatype(u'äöü', 'str')
        self.checkOutput("""
        'ascii' codec can't encode character...: ordinal not in range(128)
        """, str(arc.exception))

        # Convert to string by type
        self.assertEqual(convert_value_to_datatype(UNSET, BYTES_TYPE), UNSET)

        converted = convert_value_to_datatype(u'string', BYTES_TYPE)
        self.assertEqual(converted, b'string')
        self.assertTrue(isinstance(converted, BYTES_TYPE))

        with self.assertRaises(UnicodeEncodeError) as arc:
            convert_value_to_datatype(u'äöü', BYTES_TYPE)
        self.checkOutput("""
        'ascii' codec can't encode character...: ordinal not in range(128)
        """, str(arc.exception))

    def test_convert_value_to_datatype_unicode(self):
        # Convert to unicode by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'unicode'), UNSET)

        converted = convert_value_to_datatype(b'unicode', 'unicode')
        self.assertEqual(converted, u'unicode')
        self.assertTrue(isinstance(converted, UNICODE_TYPE))

        # Convert to unicode by type
        self.assertEqual(convert_value_to_datatype(UNSET, UNICODE_TYPE), UNSET)

        converted = convert_value_to_datatype(b'unicode', UNICODE_TYPE)
        self.assertEqual(converted, u'unicode')
        self.assertTrue(isinstance(converted, UNICODE_TYPE))

    def test_convert_value_to_datatype_int(self):
        # Convert to int by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'int'), UNSET)

        converted = convert_value_to_datatype('1', 'int')
        self.assertEqual(converted, 1)
        self.assertTrue(isinstance(converted, int))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('1.0', 'int')
        msg = "invalid literal for int() with base 10: '1.0'"
        self.assertEqual(str(arc.exception), msg)

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'int')
        msg = "invalid literal for int() with base 10: 'a'"
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(2.0, 'int')
        self.assertEqual(converted, 2)
        self.assertTrue(isinstance(converted, int))

        # Convert to int by type
        self.assertEqual(convert_value_to_datatype(UNSET, int), UNSET)

        converted = convert_value_to_datatype('3', int)
        self.assertEqual(converted, 3)
        self.assertTrue(isinstance(converted, int))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('2.0', int)
        msg = "invalid literal for int() with base 10: '2.0'"
        self.assertEqual(str(arc.exception), msg)

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', int)
        msg = "invalid literal for int() with base 10: 'b'"
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(4.0, int)
        self.assertEqual(converted, 4)
        self.assertTrue(isinstance(converted, int))

    def test_convert_value_to_datatype_long(self):
        # Convert to long by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'long'), UNSET)

        converted = convert_value_to_datatype('1', 'long')
        self.assertEqual(converted, LONG_TYPE(1))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        converted = convert_value_to_datatype(2.0, 'long')
        self.assertEqual(converted, LONG_TYPE(2))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'long')
        # there is no long type in python 3, falls back to int
        msg = (
            "invalid literal for long() with base 10: 'a'"
            if IS_PY2
            else "invalid literal for int() with base 10: 'a'"
        )
        self.assertEqual(str(arc.exception), msg)

        # Convert to long by type
        self.assertEqual(convert_value_to_datatype(UNSET, LONG_TYPE), UNSET)

        converted = convert_value_to_datatype('3', LONG_TYPE)
        self.assertEqual(converted, LONG_TYPE(3))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        converted = convert_value_to_datatype(4.0, LONG_TYPE)
        self.assertEqual(converted, LONG_TYPE(4))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', LONG_TYPE)
        # there is no long type in python 3, falls back to int
        msg = (
            "invalid literal for long() with base 10: 'b'"
            if IS_PY2
            else "invalid literal for int() with base 10: 'b'"
        )
        self.assertEqual(str(arc.exception), msg)

    def test_convert_value_to_datatype_float(self):
        # Convert to float by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'float'), UNSET)

        converted = convert_value_to_datatype('1,0', 'float')
        self.assertEqual(converted, 1.0)
        self.assertTrue(isinstance(converted, float))

        converted = convert_value_to_datatype('2', 'float')
        self.assertEqual(converted, 2.0)
        self.assertTrue(isinstance(converted, float))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'float')
        msg = (
            "could not convert string to float: a"
            if IS_PY2
            else "could not convert string to float: 'a'"
        )
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(3, 'float')
        self.assertEqual(converted, 3.0)
        self.assertTrue(isinstance(converted, float))

        # Convert to float by type
        self.assertEqual(convert_value_to_datatype(UNSET, float), UNSET)

        converted = convert_value_to_datatype('4,0', float)
        self.assertEqual(converted, 4.0)
        self.assertTrue(isinstance(converted, float))

        converted = convert_value_to_datatype('5', float)
        self.assertEqual(converted, 5.0)
        self.assertTrue(isinstance(converted, float))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', float)
        msg = (
            "could not convert string to float: b"
            if IS_PY2
            else "could not convert string to float: 'b'"
        )
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(6, float)
        self.assertEqual(converted, 6.0)
        self.assertTrue(isinstance(converted, float))

    def test_convert_value_to_datatype_uuid(self):
        # Convert to uuid by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'uuid'), UNSET)

        converted = convert_value_to_datatype(str(uuid.uuid4()), 'uuid')
        self.assertTrue(isinstance(converted, uuid.UUID))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'uuid')
        msg = 'badly formed hexadecimal UUID string'
        self.assertEqual(str(arc.exception), msg)

        # Convert to uuid by type
        self.assertEqual(convert_value_to_datatype(UNSET, uuid.UUID), UNSET)

        converted = convert_value_to_datatype(str(uuid.uuid4()), uuid.UUID)
        self.assertTrue(isinstance(converted, uuid.UUID))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', uuid.UUID)
        msg = 'badly formed hexadecimal UUID string'
        self.assertEqual(str(arc.exception), msg)

    def test_convert_value_to_datatype_function(self):
        # Custom converter as function
        def convert_func(val):
            if val == 'a':
                return 'convertet: {0}'.format(val)
            raise ValueError("Value not 'a'")

        self.assertEqual(
            convert_value_to_datatype('a', convert_func),
            'convertet: a'
        )

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', convert_func)
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_value_to_datatype_class(self):
        # Custom converters as class
        class Converter(object):

            def __init__(self, val):
                if val != 'a':
                    raise ValueError("Value not 'a'")

        converted = convert_value_to_datatype('a', Converter)
        self.assertTrue(isinstance(converted, Converter))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', Converter)
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_value_to_datatype_instance(self):
        # Custom converter as class instance with ``__call__`` function
        class ConverterInst(object):

            def __call__(self, val):
                if val != 'a':
                    raise ValueError("Value not 'a'")
                return 'convertet: {0}'.format(val)

        self.assertEqual(
            convert_value_to_datatype('a', ConverterInst()),
            'convertet: a'
        )

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', ConverterInst())
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_values_to_datatype(self):
        self.assertEqual(convert_values_to_datatype(UNSET, 'int'), UNSET)
        self.assertEqual(convert_values_to_datatype([UNSET], 'int'), [UNSET])
        self.assertEqual(convert_values_to_datatype('0', int), 0)
        self.assertEqual(convert_values_to_datatype(['0', '1'], int), [0, 1])

    def test_datatype_extractor(self):
        # No datatype given, no datatype conversion happens at all
        widget = factory(
            'text',
            name='MYFIELD',
            value='')
        data = widget.extract({'MYFIELD': u''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, '')

        # Test emptyvalue if ``str`` datatype set
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'str',
            })

        # Default emptyvalue
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # None emptyvalue
        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, None)

        # UNSET emptyvalue
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # String emptyvalue
        widget.attrs['emptyvalue'] = 'abc'
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, b'abc')

        # Unicode emptyvalue
        widget.attrs['emptyvalue'] = u''
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # Test emptyvalue if ``int`` datatype set
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'int',
            })

        # Default emptyvalue
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # None emptyvalue
        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, None)

        # UNSET emptyvalue
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # Int emptyvalue
        widget.attrs['emptyvalue'] = -1
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, -1)

        # String emptyvalue. If convertable still fine
        widget.attrs['emptyvalue'] = '0'
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 0)

        # Test emptyvalue if ``long`` datatype set
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'long',
            })

        # Default emptyvalue
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # None emptyvalue
        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, None)

        # UNSET emptyvalue
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # Int emptyvalue
        widget.attrs['emptyvalue'] = -1
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, LONG_TYPE(-1))

        # String emptyvalue. If convertable still fine
        widget.attrs['emptyvalue'] = '0'
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, LONG_TYPE(0))

        # Test emptyvalue if ``float`` datatype set
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'float',
            })

        # Default emptyvalue
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # None emptyvalue
        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, None)

        # UNSET emptyvalue
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # Float emptyvalue
        widget.attrs['emptyvalue'] = 0.1
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 0.1)

        # String emptyvalue. If convertable still fine
        widget.attrs['emptyvalue'] = '0,2'
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 0.2)

        # Test emptyvalue if ``uuid`` datatype set
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'uuid',
            })

        # Default emptyvalue
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # None emptyvalue
        widget.attrs['emptyvalue'] = None
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, None)

        # UNSET emptyvalue
        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        # UUID emptyvalue
        uid = uuid.uuid4()
        widget.attrs['emptyvalue'] = uid
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, uid)

        # String emptyvalue. If convertable still fine
        widget.attrs['emptyvalue'] = str(uid)
        data = widget.extract({})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract({'MYDATATYPEFIELD': ''})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, uid)

        # Integer datatype
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'int',
            })
        data = widget.extract({'MYDATATYPEFIELD': '1'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 1)

        data = widget.extract({'MYDATATYPEFIELD': 'a'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input is not a valid integer.')]
        )

        # Float extraction
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'float',
            })
        data = widget.extract({'MYDATATYPEFIELD': '1.2'})
        self.assertEqual(data.errors, [])
        self.assertEqual(data.extracted, 1.2)

        data = widget.extract({'MYDATATYPEFIELD': 'a'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input is not a valid floating point number.')]
        )

        # UUID extraction
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'uuid',
            })
        data = widget.extract({
            'MYDATATYPEFIELD': '3b8449f3-0456-4baa-a670-3066b0fcbda0'
        })
        self.assertEqual(data.errors, [])
        self.assertEqual(
            data.extracted,
            uuid.UUID('3b8449f3-0456-4baa-a670-3066b0fcbda0')
        )

        data = widget.extract({'MYDATATYPEFIELD': 'a'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input is not a valid UUID.')]
        )

        # Test ``datatype`` not allowed
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'uuid',
                'allowed_datatypes': [int],
            })

        request = {
            'MYDATATYPEFIELD': '3b8449f3-0456-4baa-a670-3066b0fcbda0'
        }
        with self.assertRaises(ValueError) as arc:
            widget.extract(request)
        self.assertEqual(str(arc.exception), 'Datatype not allowed: "uuid"')

        # Test ``datatype_message``
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': int,
                'datatype_message': 'This did not work'
            })
        request = {
            'MYDATATYPEFIELD': 'a'
        }
        data = widget.extract(request)
        self.assertEqual(data.errors, [ExtractionError('This did not work')])
        self.assertEqual(data.extracted, 'a')

        # Test default error message if custom converter given but no
        # ``datatype_message`` defined
        def custom_converter(val):
            raise ValueError

        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': custom_converter,
            })
        request = {
            'MYDATATYPEFIELD': 'a'
        }
        data = widget.extract(request)
        self.assertEqual(
            data.errors,
            [ExtractionError('Input conversion failed.')]
        )
        self.assertEqual(data.extracted, 'a')

        # Test unknown string ``datatype`` identifier
        widget = factory(
            'text',
            name='MYDATATYPEFIELD',
            value='',
            props={
                'datatype': 'inexistent',
            })
        with self.assertRaises(ValueError) as arc:
            widget.extract({'MYDATATYPEFIELD': 'a'})
        self.assertEqual(str(arc.exception), 'Datatype unknown: "inexistent"')



# -*- coding: utf-8 -*-
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_generic_renderer
from yafowil.compat import BYTES_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.tsf import _
import re


###############################################################################
# email
###############################################################################

EMAIL_RE = r'^[a-zA-Z0-9\._\-]+@[a-zA-Z0-9\._\-]+.[a-zA-Z0-9]{2,6}$'


def email_extractor(widget, data):
    val = data.extracted
    if not val:
        return val
    email_re = EMAIL_RE \
        if isinstance(val, UNICODE_TYPE) \
        else EMAIL_RE.encode()
    if not re.match(email_re, val):
        raise ExtractionError(_(
            'email_address_not_valid',
            default=u'Input not a valid email address.'
        ))
    return val


factory.register(
    'email',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
        email_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['email'] = """\
Email (HTML5) input blueprint.
"""

factory.defaults['email.type'] = 'email'

factory.defaults['email.default'] = ''

factory.defaults['email.required_class'] = 'required'

factory.defaults['email.class'] = 'email'

factory.defaults['email.persist'] = True

factory.defaults['email.allowed_datatypes'] = [
    BYTES_TYPE, UNICODE_TYPE,
]



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.tests import wrapped_fxml


class TestField(YafowilTestCase):

    def test_field_blueprint(self):
        # Chained file inside field with label
        widget = factory(
            'field:label:file',
            name='MYFILE',
            props={
                'label': 'MY FILE'
            })
        self.checkOutput("""
        <div class="field" id="field-MYFILE">
          <label for="input-MYFILE">MY FILE</label>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
        </div>
        """, fxml(widget()))

        # Render error class directly on field
        widget = factory(
            'field:text',
            name='MYFIELD',
            props={
                'required': True,
                'witherror': 'fielderrorclass'
            })
        data = widget.extract({'MYFIELD': ''})
        self.assertEqual(data.name, 'MYFIELD')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('Mandatory field was empty')]
        )

        self.checkOutput("""
        <div class="field fielderrorclass" id="field-MYFIELD">
          <input class="required text" id="input-MYFIELD" name="MYFIELD"
                 required="required" type="text" value=""/>
        </div>
        """, fxml(widget(data)))

    def test_label_blueprint(self):
        # Default
        widget = factory(
            'label:file',
            name='MYFILE',
            props={
                'label': 'MY FILE'
            })
        self.checkOutput("""
        <div>
          <label for="input-MYFILE">MY FILE</label>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
        </div>
        """, wrapped_fxml(widget()))

        # Label after widget
        widget = factory(
            'label:file',
            name='MYFILE',
            props={
                'label': 'MY FILE',
                'label.position': 'after'
            })
        self.checkOutput("""
        <div>
          <input id="input-MYFILE" name="MYFILE" type="file"/>
          <label for="input-MYFILE">MY FILE</label>
        </div>
        """, wrapped_fxml(widget()))

        # Same with inner label
        widget = factory(
            'label:file',
            name='MYFILE',
            props={
                'label': 'MY FILE',
                'label.position': 'inner'
            })
        self.checkOutput("""
        <div>
          <label for="input-MYFILE">MY FILE<input id="input-MYFILE"
                 name="MYFILE" type="file"/></label>
        </div>
        """, wrapped_fxml(widget()))

        # Invalid position
        widget = factory(
            'label:file',
            name='MYFILE',
            props={
                'label': 'MY FILE',
                'label.position': 'inexistent'
            })
        with self.assertRaises(ValueError) as arc:
            widget()
        self.assertEqual(
            str(arc.exception),
            'Invalid value for position "inexistent"'
        )

        # Render with title attribute
        widget = factory(
            'label',
            name='MYFILE',
            props={
                'title': 'My awesome title',
            })
        self.assertEqual(widget(), (
            '<label for="input-MYFILE" title="My awesome title">MYFILE</label>'
        ))

        # Label Text can be a callable
        widget = factory(
            'label',
            name='MYFILE',
            props={
                'label': lambda x, y: 'Fooo',
            })
        self.assertEqual(widget(), (
            '<label for="input-MYFILE">Fooo</label>'
        ))

        # Position can be callable
        widget = factory(
            'label',
            name='MYFILE',
            props={
                'label': 'Fooo',
                'position': lambda x, y: 'inner',
            })
        self.assertEqual(widget(), '<label for="input-MYFILE">Fooo</label>')

    def test_help_blueprint(self):
        # Render some additional help text
        widget = factory(
            'field:help:text',
            name='HELPEXAMPLE',
            props={
                'label': 'Help',
                'help': 'Shout out loud here'
            })
        self.checkOutput("""
        <div class="field" id="field-HELPEXAMPLE">
          <div class="help">Shout out loud here</div>
          <input class="text" id="input-HELPEXAMPLE" name="HELPEXAMPLE"
                 type="text" value=""/>
        </div>
        """, fxml(widget()))

        # Render empty (WHAT'S THIS GOOD FOR?)
        widget = factory(
            'field:help:text',
            name='HELPEXAMPLE',
            props={
                'label': 'Help',
                'help': False,
                'render_empty': False
            })
        self.checkOutput("""
        <div class="field" id="field-HELPEXAMPLE">
          <input class="text" id="input-HELPEXAMPLE" name="HELPEXAMPLE"
                 type="text" value=""/>
        </div>
        """, fxml(widget()))

    def test_error_blueprint(self):
        # Chained password inside error inside field
        widget = factory(
            'field:error:password',
            name='PASSWORD',
            props={
                'label': 'Password',
                'required': 'No password given!'
            })
        data = widget.extract({'PASSWORD': ''})
        self.checkOutput("""
        <div class="field" id="field-PASSWORD">
          <div class="error">
            <div class="errormessage">No password given!</div>
            <input class="password required" id="input-PASSWORD" name="PASSWORD"
                   required="required" type="password" value=""/>
          </div>
        </div>
        """, fxml(widget(data=data)))  # noqa

        data = widget.extract({'PASSWORD': 'secret'})
        self.checkOutput("""
        <div class="field" id="field-PASSWORD">
          <input class="password required" id="input-PASSWORD" name="PASSWORD"
                 required="required" type="password" value="secret"/>
        </div>
        """, fxml(widget(data=data)))

        widget = factory(
            'error:text',
            name='MYDISPLAY',
            value='somevalue',
            mode='display')
        self.assertEqual(widget(), (
            '<div class="display-text" id="display-MYDISPLAY">somevalue</div>'
        ))

        # Error wrapping in div element can be suppressed
        widget = factory(
            'field:error:password',
            name='PASSWORD',
            props={
                'label': 'Password',
                'required': 'No password given!',
                'message_tag': None
            })
        data = widget.extract({'PASSWORD': ''})
        self.checkOutput("""
        <div class="field" id="field-PASSWORD">
          <div class="error">No password given!<input class="password required"
               id="input-PASSWORD" name="PASSWORD" required="required"
               type="password" value=""/></div>
        </div>
        """, fxml(widget(data=data)))



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.compat import BYTES_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestEmail(YafowilTestCase):

    def test_email_blueprint(self):
        # Render email input field
        widget = factory(
            'email',
            name='EMAIL')
        self.assertEqual(widget(), (
            '<input class="email" id="input-EMAIL" name="EMAIL" '
            'type="email" value="" />'
        ))

        # Extract not required and empty
        data = widget.extract({'EMAIL': ''})
        self.assertEqual(data.errors, [])

        # Extract invalid email input
        data = widget.extract({'EMAIL': 'foo@bar'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input not a valid email address.')]
        )

        data = widget.extract({'EMAIL': '@bar.com'})
        self.assertEqual(
            data.errors,
            [ExtractionError('Input not a valid email address.')]
        )

        # Extract valid email input
        data = widget.extract({'EMAIL': 'foo@bar.com'})
        self.assertEqual(data.errors, [])

        # Extract required email input
        widget = factory(
            'email',
            name='EMAIL',
            props={
                'required': 'E-Mail Address is required'
            })
        data = widget.extract({'EMAIL': ''})
        self.assertEqual(
            data.errors,
            [ExtractionError('E-Mail Address is required')]
        )

        data = widget.extract({'EMAIL': 'foo@bar.com'})
        self.assertEqual(data.errors, [])

        # Emptyvalue
        widget = factory(
            'email',
            name='EMAIL',
            props={
                'emptyvalue': 'foo@bar.baz',
            })
        data = widget.extract(request={'EMAIL': ''})
        self.assertEqual(data.name, 'EMAIL')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'foo@bar.baz')
        self.assertEqual(data.errors, [])

        data = widget.extract(request={'EMAIL': 'foo@baz.bam'})
        self.assertEqual(data.name, 'EMAIL')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'foo@baz.bam')
        self.assertEqual(data.errors, [])

        # Datatype
        widget = factory(
            'email',
            name='EMAIL',
            props={
                'datatype': UNICODE_TYPE
            })
        data = widget.extract(request={'EMAIL': u'foo@example.com'})
        self.assertEqual(data.extracted, u'foo@example.com')
        self.assertTrue(isinstance(data.extracted, UNICODE_TYPE))
        self.assertFalse(isinstance(data.extracted, BYTES_TYPE))

        widget = factory(
            'email',
            name='EMAIL',
            props={
                'datatype': BYTES_TYPE
            })
        data = widget.extract(request={'EMAIL': u'foo@example.com'})
        self.assertEqual(data.extracted, b'foo@example.com')
        self.assertFalse(isinstance(data.extracted, UNICODE_TYPE))
        self.assertTrue(isinstance(data.extracted, BYTES_TYPE))

        # Persist
        widget = factory(
            'email',
            name='EMAIL')
        data = widget.extract({'EMAIL': 'foo@bar.baz'})
        model = dict()
        data.persist_writer = write_mapping_writer
        data.write(model)
        self.assertEqual(model, {'EMAIL': 'foo@bar.baz'})



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.common import generic_required_extractor
from yafowil.common import input_attributes_common
from yafowil.tsf import _
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops
from yafowil.utils import vocabulary


###############################################################################
# file
###############################################################################

def file_extractor(widget, data):
    """Return a dict with following keys:

    mimetype
        Mimetype of file.
    headers
        rfc822.Message instance.
    original
        Original file handle from underlying framework.
    file
        File descriptor containing the data.
    filename
        File name.
    action
        widget flags 'new', 'keep', 'replace', 'delete'
    """
    name = widget.dottedpath
    if name not in data.request:
        return UNSET
    if not '{0}-action'.format(name) in data.request:
        value = data.request[name]
        if value:
            value['action'] = 'new'
        return value
    value = data.value
    action = value['action'] = data.request.get(
        '{0}-action'.format(name),
        'keep'
    )
    if action == 'delete':
        value['file'] = UNSET
    elif action == 'replace':
        new_val = data.request[name]
        if not new_val:
            raise ExtractionError(_(
                'file_replace_no_upload',
                default='Cannot replace file. No file uploaded.'
            ))
        value = new_val
        value['action'] = 'replace'
    return value


@managedprops('accept')
def mimetype_extractor(widget, data):
    accept = attr_value('accept', widget, data)
    extracted = data.extracted
    if not extracted or not accept:
        return extracted
    extracted_mimetype = extracted.get('mimetype')
    if not extracted_mimetype:
        return extracted
    extracted_type, extracted_sub = extracted_mimetype.split('/')
    matches = False
    for mimetype in accept.split(','):
        type_, sub = mimetype.split('/')
        if type_ == '*':
            matches = True
            break
        if type_ != extracted_type:
            continue
        if sub == '*':
            matches = True
            break
        if sub == extracted_sub:
            matches = True
            break
    if not matches:
        message = _(
            'file_invalid_mimetype',
            default=u'Mimetype of uploaded file not matches'
        )
        raise ExtractionError(message)
    return extracted


@managedprops(
    'accept',
    'placeholder',
    'autofocus',
    'required',
    *css_managed_props)
def input_file_edit_renderer(widget, data):
    tag = data.tag
    input_attrs = input_attributes_common(widget, data, excludes=['value'])
    input_attrs['type'] = 'file'
    if attr_value('accept', widget, data):
        input_attrs['accept'] = attr_value('accept', widget, data)
    return tag('input', **input_attrs)


def convert_bytes(value):
    value = float(value)
    if value >= 1099511627776:
        terabytes = value / 1099511627776
        size = '{0:.2f}T'.format(terabytes)
    elif value >= 1073741824:
        gigabytes = value / 1073741824
        size = '{0:.2f}G'.format(gigabytes)
    elif value >= 1048576:
        megabytes = value / 1048576
        size = '{0:.2f}M'.format(megabytes)
    elif value >= 1024:
        kilobytes = value / 1024
        size = '{0:.2f}K'.format(kilobytes)
    else:
        size = '{0:.2f}b'.format(value)
    return size


def input_file_display_renderer(widget, data):
    tag = data.tag
    value = data.value
    attrs = {
        'class': cssclasses(widget, data),
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    if not value:
        no_file_message = _('no_file', default=u'No file')
        return tag('div', no_file_message, **attrs)
    file_val = value['file']
    size = convert_bytes(len(file_val.read()))
    file_val.seek(0)
    unknown_message = _('unknown', default=u'Unknown')
    filename_message = _('filename', default=u'Filename: ')
    mimetype_message = _('mimetype', default=u'Mimetype: ')
    size_message = _('size', default=u'Size: ')
    filename = value.get('filename', unknown_message)
    mimetype = value.get('mimetype', unknown_message)
    return tag(
        'div',
        tag(
            'ul',
            tag('li', tag('strong', filename_message), filename),
            tag('li', tag('strong', mimetype_message), mimetype),
            tag('li', tag('strong', size_message), size)),
        **attrs
    )


@managedprops('vocabulary', *css_managed_props)
def file_options_renderer(widget, data):
    if data.value in [None, UNSET, '']:
        return data.rendered
    tag = data.tag
    if data.request:
        value = [
            data.request.get('{0}-action'.format(widget.dottedpath), 'keep')
        ]
    else:
        value = ['keep']
    tags = []
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        attrs = {
            'type': 'radio',
            'value': key,
            'checked': (key in value) and 'checked' or None,
            'name_': '{0}-action'.format(widget.dottedpath),
            'id': cssid(widget, 'input', key),
            'class_': cssclasses(widget, data),
        }
        taginput = tag('input', **attrs)
        text = tag('span', term)
        tags.append(tag(
            'div',
            taginput,
            text,
            **{'id': cssid(widget, 'radio', key)}
        ))
    return data.rendered + u''.join(tags)


factory.register(
    'file',
    extractors=[
        file_extractor,
        mimetype_extractor,
        generic_required_extractor
    ],
    edit_renderers=[
        input_file_edit_renderer,
        file_options_renderer
    ],
    display_renderers=[input_file_display_renderer]
)

factory.doc['blueprint']['file'] = """\
A basic file upload blueprint.
"""

factory.defaults['file.accept'] = None
factory.doc['props']['file.accept'] = """\
The accept attribute value is a string that defines the file types the file
input should accept. This string is a comma-separated list of unique file type
specifiers. Because a given file type may be identified in more than one
manner, it's useful to provide a thorough set of type specifiers when you need
files of a given format.
"""

factory.defaults['file.vocabulary'] = [
    ('keep', _('file_keep', default=u'Keep Existing file')),
    ('replace', _('file_replace', default=u'Replace existing file')),
    ('delete', _('file_delete', default=u'Delete existing file')),
]
factory.doc['props']['file.vocabulary'] = """\
Vocabulary with available actions for existing files.
"""



from yafowil.base import factory
from yafowil.tests import YafowilTestCase


class TestTag(YafowilTestCase):

    def test_tag_blueprint(self):
        # Custom tag widget
        widget = factory(
            'tag',
            name='MYTAG',
            props={
                'tag': 'h3',
                'text': 'A Headline',
                'class': 'form_heading'
            })
        self.assertEqual(
            widget(),
            '<h3 class="form_heading" id="tag-MYTAG">A Headline</h3>'
        )

        # Skip tag
        widget = factory(
            'tag',
            name='MYTAG',
            props={
                'tag': 'h3',
                'text': 'A Headline',
                'class': 'form_heading'
            },
            mode='skip')
        self.assertEqual(widget(), '')



from node.utils import UNSET
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
from yafowil.utils import EMPTY_VALUE


class TestHidden(YafowilTestCase):

    def test_hidden_blueprint(self):
        # Hidden input blueprint
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            value='Test Hidden')
        self.assertEqual(widget(), (
            '<input class="hidden" id="input-MYHIDDEN" name="MYHIDDEN" '
            'type="hidden" value="Test Hidden" />'
        ))

        # Display mode of hidden widget renders empty string
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            value='Test Hidden',
            mode='display')
        self.assertEqual(widget(), '')

        # As well does skip mode
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            value='Test Hidden',
            mode='skip')
        self.assertEqual(widget(), '')

        # Generic HTML5 Data
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            value='Test Hidden',
            props={
                'data': {
                    'foo': 'bar'
                }
            })
        self.assertEqual(widget(), (
            '<input class="hidden" data-foo=\'bar\' id="input-MYHIDDEN" '
            'name="MYHIDDEN" type="hidden" value="Test Hidden" />'
        ))

        # Emptyvalue
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            props={
                'emptyvalue': 'EMPTYVALUE'
            })
        data = widget.extract(request={'MYHIDDEN': ''})
        self.assertEqual(data.name, 'MYHIDDEN')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 'EMPTYVALUE')

        # Datatype
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            props={
                'emptyvalue': 0,
                'datatype': int
            })
        data = widget.extract(request={'MYHIDDEN': ''})
        self.assertEqual(data.name, 'MYHIDDEN')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, 0)

        # Default emptyvalue extraction
        del widget.attrs['emptyvalue']
        data = widget.extract(request={'MYHIDDEN': ''})
        self.assertEqual(data.name, 'MYHIDDEN')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, EMPTY_VALUE)

        # Persist property
        widget = factory(
            'hidden',
            name='MYHIDDEN',
            props={
                'emptyvalue': 0,
                'datatype': int,
                'persist_writer': write_mapping_writer,
            })
        data = widget.extract(request={'MYHIDDEN': '10'})
        model = dict()
        data.write(model)
        self.assertEqual(model, {'MYHIDDEN': 10})

        data.persist_target = 'myhidden'
        model = dict()
        data.write(model)
        self.assertEqual(model, {'myhidden': 10})



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops


###############################################################################
# textarea
###############################################################################

def textarea_attributes(widget, data):
    autofocus = attr_value('autofocus', widget, data) and 'autofocus' or None
    disabled = attr_value('disabled', widget, data) and 'disabled' or None
    readonly = attr_value('readonly', widget, data) and 'readonly' or None
    required = attr_value('required', widget, data) and 'required' or None
    ta_attrs = {
        'autofocus': autofocus,
        'class_': cssclasses(widget, data),
        'cols': attr_value('cols', widget, data),
        'disabled': disabled,
        'id': cssid(widget, 'input'),
        'title': attr_value('title', widget, data),
        'name_': widget.dottedpath,
        'placeholder': attr_value('placeholder', widget, data),
        'readonly': readonly,
        'required': required,
        'rows': attr_value('rows', widget, data),
        'wrap': attr_value('wrap', widget, data),
    }
    ta_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    return ta_attrs


textarea_managed_props = [
    'data', 'title', 'autofocus', 'cols', 'disabled',
    'placeholder', 'readonly', 'required', 'rows', 'wrap',
] + css_managed_props


@managedprops(*textarea_managed_props)
def textarea_renderer(widget, data, custom_attrs={}):
    """Renders text area.
    """
    tag = data.tag
    area_attrs = textarea_attributes(widget, data)
    area_attrs.update(custom_attrs)
    value = fetch_value(widget, data)
    if value is None:
        value = ''
    return tag('textarea', value, **area_attrs)


factory.register(
    'textarea',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor
    ],
    edit_renderers=[textarea_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['textarea'] = """\
HTML textarea blueprint.
"""

factory.defaults['textarea.default'] = ''

factory.defaults['textarea.wrap'] = None
factory.doc['props']['textarea.wrap'] = """\
Either ``soft``, ``hard``, ``virtual``, ``physical`` or  ``off``.
"""

factory.defaults['textarea.class'] = 'textarea'

factory.defaults['textarea.cols'] = 80
factory.doc['props']['textarea.cols'] = """\
Number of characters.
"""

factory.defaults['textarea.rows'] = 25
factory.doc['props']['textarea.rows'] = """\
Number of lines.
"""

factory.defaults['textarea.readonly'] = None
factory.doc['props']['textarea.readonly'] = """\
Flag textarea is readonly.
"""

factory.defaults['textarea.persist'] = True



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.common import empty_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import input_generic_renderer
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor


###############################################################################
# hidden
###############################################################################

factory.register(
    'hidden',
    extractors=[
        generic_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[empty_display_renderer]
)

factory.doc['blueprint']['hidden'] = """\
Hidden input blueprint.
"""

factory.defaults['hidden.type'] = 'hidden'
factory.doc['props']['hidden.type'] = """\
Type of input tag.
"""

factory.defaults['hidden.default'] = ''

factory.defaults['hidden.class'] = 'hidden'

factory.defaults['hidden.persist'] = True



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.compat import ITER_TYPES
from yafowil.compat import STR_TYPE
from yafowil.datatypes import lookup_datatype_converter
from yafowil.tsf import _
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops
from zope.deferredimport import deprecated


###############################################################################
# common defaults
###############################################################################

factory.defaults['default'] = UNSET
factory.doc['props']['default'] = """\
Default value for rendering.
"""

factory.defaults['class'] = None
factory.doc['props']['class'] = """\
Common CSS-class to put on.
"""

factory.defaults['class_add'] = None
factory.doc['props']['class_add'] = """\
Additional CSS-class to put on.
"""

factory.defaults['error_class'] = None
factory.doc['props']['error_class'] = """\
CSS-class to put on in case of error after extraction.
"""

factory.defaults['error_class_default'] = 'error'
factory.doc['props']['error_class_default'] = """\
Fallback CSS-class to put on in case of error if no specific class was
given.
"""

factory.defaults['valid_class'] = None
factory.doc['props']['valid_class'] = """\
CSS-class to put on in case of valid value after extraction.
"""

factory.defaults['valid_class_default'] = 'valid'
factory.doc['props']['valid_class_default'] = """\
Fallback CSS-class to put on in case of valid value if no specific class was
given.
"""

factory.defaults['autofocus'] = None
factory.doc['props']['autofocus'] = """\
Whether this field gets the focus automatically or not (if browser supports
it).
"""

factory.defaults['autocomplete'] = None
factory.doc['props']['autocomplete'] = """\
Switch autocomplete explizit to ``on`` or ``off``.
"""

factory.defaults['placeholder'] = None
factory.doc['props']['placeholder'] = """\
Whether this input has a placeholder value or not (if browser supports it).
"""

factory.defaults['required'] = False
factory.doc['props']['required'] = """\
Whether this value is required or not.
"""

factory.defaults['required_message'] = _(
    'required_message', default=u'Mandatory field was empty')
factory.doc['props']['required_message'] = """\
Message to be shown if required condition was not met.
"""

factory.defaults['required_class'] = None
factory.doc['props']['required_class'] = """\
CSS-class to put on in case if required condition was not met.
"""

factory.defaults['type'] = None
factory.doc['props']['type'] = """\
HTML type attribute.
"""

factory.defaults['size'] = None
factory.doc['props']['size'] = """\
Allowed input size.
"""

factory.defaults['maxlength'] = None
factory.doc['props']['maxlength'] = """\
Input maxlength.
"""

factory.defaults['disabled'] = None
factory.doc['props']['disabled'] = """\
Disables input.
"""

factory.defaults['required_class_default'] = 'required'
factory.doc['props']['required_class_default'] = """\
CSS-class to apply if required condition was not met - if no specific class
was given.
"""

factory.defaults['template'] = '%s'
factory.doc['props']['template'] = """\
Format string with pythons built-in string format template. If a callable
is given it will be used instead and is called with ``widget`` and ``data`` as
parameters.
"""

factory.defaults['title'] = None
factory.doc['props']['title'] = """\
Optional help text to be rendered in the title attribute.
"""

factory.defaults['data'] = dict()
factory.doc['props']['data'] = """\
Additional data rendered as HTML5 data attributes on DOM Element.
"""

factory.defaults['display_proxy'] = False
factory.doc['props']['display_proxy'] = """\
If 'True' and widget mode 'display', widget value gets rendered as hidden
input.
"""


###############################################################################
# generic
###############################################################################

def generic_extractor(widget, data):
    """Extract raw value from request by ``widget.dottedpath``.

    Return ``UNSET`` if widget not present in request.
    """
    __managed_props = []  # noqa
    try:
        return data.request[widget.dottedpath]
    except KeyError:
        return UNSET


@managedprops('required', 'required_message')
def generic_required_extractor(widget, data):
    """Validate required.

    If ``required`` is set and some raw value has been extracted,
    evaluate ``data.extracted`` to boolean. Raise ``ExtractionError`` if
    ``False``.
    """
    required = attr_value('required', widget, data)
    if not required or bool(data.extracted) or data.extracted is UNSET:
        return data.extracted
    if isinstance(required, STR_TYPE):
        raise ExtractionError(required)
    raise ExtractionError(attr_value('required_message', widget, data))


def input_attributes_common(widget, data, excludes=list(), value=None):
    if value is None:
        value = fetch_value(widget, data)
    datatype = widget.attrs.get('datatype', None)
    if datatype is not None:
        converter = lookup_datatype_converter(datatype)
        value = converter.to_form(value)
    # XXX: get rid of this, we probaly want to have a default datatype
    #      converter here
    elif isinstance(value, STR_TYPE):
        value = value.replace('"', '&quot;')
    autofocus = attr_value('autofocus', widget, data) and 'autofocus' or None
    disabled = attr_value('disabled', widget, data)
    disabled = bool(disabled) and 'disabled' or None
    required = attr_value('required', widget, data) and 'required' or None
    input_attrs = {
        'autofocus': autofocus,
        'class_': cssclasses(widget, data),
        'disabled': disabled,
        'id': cssid(widget, 'input'),
        'name_': widget.dottedpath,
        'placeholder': attr_value('placeholder', widget, data),
        'required': required,
        'size': attr_value('size', widget, data),
        'maxlength': attr_value('maxlength', widget, data),
        'title': attr_value('title', widget, data),
        'type': attr_value('type', widget, data),
        'value': value,
    }
    input_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    for attr_name in excludes:
        del input_attrs[attr_name]
    return input_attrs


def input_attributes_full(widget, data, value=None):
    input_attrs = input_attributes_common(widget, data, value=value)
    input_attrs['autocomplete'] = attr_value(
        'autocomplete', widget, data)
    if attr_value('type', widget, data) in ['range', 'number']:
        input_attrs['min'] = attr_value('min', widget, data)
        input_attrs['max'] = attr_value('max', widget, data)
        input_attrs['step'] = attr_value('step', widget, data)
    return input_attrs


@managedprops(*css_managed_props)
def input_generic_renderer(widget, data, custom_attrs={}):
    """Generic HTML ``input`` tag render.
    """
    input_attrs = input_attributes_full(widget, data)
    input_attrs.update(custom_attrs)
    return data.tag('input', **input_attrs)


# multivalued is not documented, because its only valid for specific blueprints
@managedprops('display_proxy')
def display_proxy_renderer(widget, data):
    rendered = data.rendered
    if attr_value('display_proxy', widget, data):
        orgin_type = attr_value('type', widget, data)
        widget.attrs['type'] = 'hidden'
        value = fetch_value(widget, data)
        multivalued = attr_value('multivalued', widget, data)
        if multivalued and isinstance(value, STR_TYPE):
            value = [value]
        if multivalued or type(value) in ITER_TYPES:
            for val in value:
                input_attrs = input_attributes_full(widget, data, value=val)
                rendered += data.tag('input', **input_attrs)
        else:
            rendered += input_generic_renderer(widget, data)
        if orgin_type:
            widget.attrs['type'] = orgin_type
        else:
            del widget.attrs['type']
    return rendered


@managedprops('template', 'class')
def generic_display_renderer(widget, data, value=None):
    """Generic display renderer to render a value.
    """
    if callable(widget.attrs['template']):
        content = widget.attrs['template'](widget, data)
    elif value is None:
        value = fetch_value(widget, data)
        if value in [UNSET, None]:
            value = u''
        content = widget.attrs['template'] % value
    else:
        content = widget.attrs['template'] % value
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(
            attr_value('class', widget, data) or 'generic'
        )
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    return data.tag('div', content, **attrs)


def empty_display_renderer(widget, data):
    """Display renderer which renders an empty string.
    """
    return data.rendered or u''


def generic_positional_rendering_helper(
        tagname, message, attrs, rendered, pos, tag
    ):
    """returns new tag with rendered content dependent on position

    tagname
        name of new tag, ie. div, p, label and so on

    message
        text included in new tag

    attrs
        attributes on new tag

    rendered
        prior rendered text

    pos
        position how to place the newtag relative to the prior rendered:
        'before'='<newtag>message</newtag>rendered',
        'after' ='<newtag>message</newtag>'
        'inner-before'= <newtag>message rendered</newtag>
        'inner-after'= <newtag>rendered message</newtag>
    """
    if pos not in ['before', 'after', 'inner-before', 'inner-after']:
        raise ValueError('Invalid value for position "{0}"'.format(pos))
    if pos.startswith('inner'):
        if pos.endswith('before'):
            inner = message, rendered
        else:
            inner = rendered, message
        return tag(tagname, *inner, **attrs)
    else:
        newtag = tag(tagname, message, **attrs)
        if pos == 'before':
            return newtag + rendered
        return rendered + newtag


###############################################################################
# B/C imports
###############################################################################

# button
deprecated(
    '``submit_renderer`` has been moved to ``yafowil.button``.',
    submit_renderer='yafowil.button:submit_renderer'
)
deprecated(
    '``button_renderer`` has been moved to ``yafowil.button``.',
    button_renderer='yafowil.button:button_renderer'
)

# checkbox
deprecated(
    '``checkbox_extractor`` has been moved to ``yafowil.checkbox``.',
    checkbox_extractor='yafowil.checkbox:checkbox_extractor'
)
deprecated(
    '``checkbox_edit_renderer`` has been moved to ``yafowil.checkbox``.',
    checkbox_edit_renderer='yafowil.checkbox:checkbox_edit_renderer'
)
deprecated(
    '``checkbox_display_renderer`` has been moved to ``yafowil.checkbox``.',
    checkbox_display_renderer='yafowil.checkbox:checkbox_display_renderer'
)

# datatypes
deprecated(
    '``generic_emptyvalue_extractor`` has been moved to ``yafowil.datatypes``.',
    generic_emptyvalue_extractor='yafowil.datatypes:generic_emptyvalue_extractor'
)
deprecated(
    '``generic_datatype_extractor`` has been moved to ``yafowil.datatypes``.',
    generic_datatype_extractor='yafowil.datatypes:generic_datatype_extractor'
)
deprecated(
    '``DATATYPE_LABELS`` has been moved to ``yafowil.datatypes``.',
    DATATYPE_LABELS='yafowil.datatypes:DATATYPE_LABELS'
)

# email
deprecated(
    '``email_extractor`` has been moved to ``yafowil.email``.',
    email_extractor='yafowil.email:email_extractor'
)

# field
deprecated(
    '``field_renderer`` has been moved to ``yafowil.field``.',
    field_renderer='yafowil.field:field_renderer'
)
deprecated(
    '``label_renderer`` has been moved to ``yafowil.field``.',
    label_renderer='yafowil.field:label_renderer'
)
deprecated(
    '``help_renderer`` has been moved to ``yafowil.field``.',
    help_renderer='yafowil.field:help_renderer'
)
deprecated(
    '``error_renderer`` has been moved to ``yafowil.field``.',
    error_renderer='yafowil.field:error_renderer'
)

# file
deprecated(
    '``file_extractor`` has been moved to ``yafowil.file``.',
    file_extractor='yafowil.file:file_extractor'
)
deprecated(
    '``mimetype_extractor`` has been moved to ``yafowil.file``.',
    mimetype_extractor='yafowil.file:mimetype_extractor'
)
deprecated(
    '``input_file_edit_renderer`` has been moved to ``yafowil.file``.',
    input_file_edit_renderer='yafowil.file:input_file_edit_renderer'
)
deprecated(
    '``convert_bytes`` has been moved to ``yafowil.file``.',
    convert_bytes='yafowil.file:convert_bytes'
)
deprecated(
    '``input_file_display_renderer`` has been moved to ``yafowil.file``.',
    input_file_display_renderer='yafowil.file:input_file_display_renderer'
)
deprecated(
    '``file_options_renderer`` has been moved to ``yafowil.file``.',
    file_options_renderer='yafowil.file:file_options_renderer'
)

# lines
deprecated(
    '``lines_extractor`` has been moved to ``yafowil.lines``.',
    lines_extractor='yafowil.lines:lines_extractor'
)
deprecated(
    '``lines_edit_renderer`` has been moved to ``yafowil.lines``.',
    lines_edit_renderer='yafowil.lines:lines_edit_renderer'
)
deprecated(
    '``lines_display_renderer`` has been moved to ``yafowil.lines``.',
    lines_display_renderer='yafowil.lines:lines_display_renderer'
)

# number
deprecated(
    '``number_extractor`` has been moved to ``yafowil.number``.',
    number_extractor='yafowil.number:number_extractor'
)

# password
deprecated(
    '``minlength_extractor`` has been moved to ``yafowil.password``.',
    minlength_extractor='yafowil.password:minlength_extractor'
)
deprecated(
    '``ascii_extractor`` has been moved to ``yafowil.password``.',
    ascii_extractor='yafowil.password:ascii_extractor'
)
deprecated(
    '``password_extractor`` has been moved to ``yafowil.password``.',
    password_extractor='yafowil.password:password_extractor'
)
deprecated(
    '``password_edit_renderer`` has been moved to ``yafowil.password``.',
    password_edit_renderer='yafowil.password:password_edit_renderer'
)
deprecated(
    '``password_display_renderer`` has been moved to ``yafowil.password``.',
    password_display_renderer='yafowil.password:password_display_renderer'
)

# proxy
deprecated(
    '``input_proxy_renderer`` has been moved to ``yafowil.proxy``.',
    input_proxy_renderer='yafowil.proxy:input_proxy_renderer'
)

# select
deprecated(
    '``select_extractor`` has been moved to ``yafowil.select``.',
    select_extractor='yafowil.select:select_extractor'
)
deprecated(
    '``select_exists_marker`` has been moved to ``yafowil.select``.',
    select_exists_marker='yafowil.select:select_exists_marker'
)
deprecated(
    '``select_edit_renderer_props`` has been moved to ``yafowil.select``.',
    select_edit_renderer_props='yafowil.select:select_edit_renderer_props'
)
deprecated(
    '``select_block_edit_renderer`` has been moved to ``yafowil.select``.',
    select_block_edit_renderer='yafowil.select:select_block_edit_renderer'
)
deprecated(
    '``select_cb_edit_renderer`` has been moved to ``yafowil.select``.',
    select_cb_edit_renderer='yafowil.select:select_cb_edit_renderer'
)
deprecated(
    '``select_edit_renderer`` has been moved to ``yafowil.select``.',
    select_edit_renderer='yafowil.select:select_edit_renderer'
)
deprecated(
    '``select_display_renderer`` has been moved to ``yafowil.select``.',
    select_display_renderer='yafowil.select:select_display_renderer'
)

# tag
deprecated(
    '``tag_renderer`` has been moved to ``yafowil.tag``.',
    tag_renderer='yafowil.tag:tag_renderer'
)

# text
deprecated(
    '``text_edit_renderer`` has been moved to ``yafowil.text``.',
    text_edit_renderer='yafowil.text:text_edit_renderer'
)

# textarea
deprecated(
    '``textarea_managed_props`` has been moved to ``yafowil.textarea``.',
    textarea_managed_props='yafowil.textarea:textarea_managed_props'
)
deprecated(
    '``textarea_attributes`` has been moved to ``yafowil.textarea``.',
    textarea_attributes='yafowil.textarea:textarea_attributes'
)
deprecated(
    '``textarea_renderer`` has been moved to ``yafowil.textarea``.',
    textarea_renderer='yafowil.textarea:textarea_renderer'
)

# url
deprecated(
    '``url_extractor`` has been moved to ``yafowil.url``.',
    url_extractor='yafowil.url:url_extractor'
)



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import factory
from yafowil.common import empty_display_renderer
from yafowil.common import generic_positional_rendering_helper
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops


###############################################################################
# field
###############################################################################

@managedprops('witherror', *css_managed_props)
def field_renderer(widget, data):
    tag = data.tag
    div_attrs = {
        'id': cssid(widget, 'field'),
        'class_': cssclasses(widget, data)
    }
    witherror = attr_value('witherror', widget, data)
    if witherror and data.errors:
        div_attrs['class_'] += u' {0}'.format(witherror)
    return tag('div', data.rendered, **div_attrs)


factory.register(
    'field',
    edit_renderers=[field_renderer],
    display_renderers=[field_renderer]
)

factory.doc['blueprint']['field'] = """\
Renders a div with an class field around the prior rendered output. This is
supposed to be used for styling and grouping purposes.
"""

factory.defaults['field.class'] = 'field'

factory.defaults['field.witherror'] = None
factory.doc['props']['field.witherror'] = """\
Put the class given with this property on the div if an error happened.
"""


###############################################################################
# label
###############################################################################

@managedprops('position', 'label', 'for', *css_managed_props)
def label_renderer(widget, data):
    tag = data.tag
    label_text = attr_value('label', widget, data, widget.name)
    label_attrs = {
        'class_': cssclasses(widget, data)
    }
    if data.mode == 'edit':
        for_path = attr_value('for', widget, data)
        if for_path:
            for_widget = widget.root
            for name in for_path.split('.'):
                for_widget = for_widget[name]
            label_attrs['for_'] = cssid(for_widget, 'input')
        else:
            label_attrs['for_'] = cssid(widget, 'input')
        title = attr_value('title', widget, data)
        if title:
            label_attrs['title'] = title
    pos = attr_value('position', widget, data)
    if pos == 'inner':
        # deprecated, use explicit inner-after or inner-before
        pos = 'inner-before'
    rendered = data.rendered is not UNSET and data.rendered or u''
    return generic_positional_rendering_helper(
        'label', label_text, label_attrs, rendered, pos, tag
    )


factory.register(
    'label',
    edit_renderers=[label_renderer],
    display_renderers=[label_renderer]
)

factory.doc['blueprint']['label'] = """\
Label blueprint.
"""

factory.defaults['label.position'] = 'before'
factory.doc['props']['label.position'] = """\
Label can be rendered at 3 different positions: ``before`` or ``after`` the
prior rendered output or with ``inner`` it puts the prior rendered output
inside the label tag.
"""

factory.doc['props']['label.label'] = """\
Text to be displayed as a label.
"""

factory.defaults['label.for'] = None
factory.doc['props']['label.for'] = """\
Optional dottedpath of widget to be labled
"""


###############################################################################
# help
###############################################################################

@managedprops('tag', 'help', 'position', 'render_empty', *css_managed_props)
def help_renderer(widget, data):
    render_empty = attr_value('render_empty', widget, data)
    help_val = attr_value('help', widget, data)
    if not render_empty and not help_val:
        return data.rendered
    tag = data.tag
    attrs = dict(class_=cssclasses(widget, data))
    elem_tag = attr_value('tag', widget, data)
    position = attr_value('position', widget, data)
    return generic_positional_rendering_helper(
        elem_tag, help_val, attrs, data.rendered, position, tag
    )


factory.register(
    'help',
    edit_renderers=[help_renderer],
    display_renderers=[empty_display_renderer]
)

factory.doc['blueprint']['help'] = """\
Renders a tag with an help-message and the prior rendered output.
"""

factory.defaults['help.class'] = 'help'

factory.defaults['help.tag'] = 'div'
factory.doc['props']['help.tag'] = """\
HTML tag to use to enclose all help messages.
"""

factory.defaults['help.help'] = ''
factory.doc['props']['help.help'] = """\
Help text.
"""

factory.defaults['help.render_empty'] = False
factory.doc['props']['help.render_empty'] = """\
Render tag even if there is no help message.
"""

factory.defaults['help.position'] = 'before'
factory.doc['props']['help.position'] = """\
Help can be rendered at 3 different positions: ``before``/ ``after`` the
prior rendered output or with ``inner-before``/ ``inner-after``  it puts the
prior rendered output inside the tag used for the help message (beofre or
after the message.
"""


###############################################################################
# error
###############################################################################

@managedprops(
    'tag',
    'message_tag',
    'message_class',
    'position',
    'render_empty',
    *css_managed_props)
def error_renderer(widget, data):
    if not data.errors and not attr_value('render_empty', widget, data):
        return data.rendered
    tag = data.tag
    msgs = u''
    for error in data.errors:
        message_tag = attr_value('message_tag', widget, data)
        if message_tag:
            msgs += tag(
                message_tag,
                error.msg,
                class_=attr_value('message_class', widget, data)
            )
        else:
            msgs += tag.translate(error.msg)
    attrs = dict(class_=cssclasses(widget, data))
    elem_tag = attr_value('tag', widget, data)
    position = attr_value('position', widget, data)
    return generic_positional_rendering_helper(
        elem_tag, msgs, attrs, data.rendered, position, tag
    )


factory.register(
    'error',
    edit_renderers=[error_renderer],
    display_renderers=[empty_display_renderer]
)

factory.doc['blueprint']['error'] = """\
Renders a tag with an error-message and the prior rendered output.
"""

factory.defaults['error.class'] = 'error'

factory.defaults['error.tag'] = 'div'
factory.doc['props']['error.tag'] = """\
HTML tag to use to enclose all error messages.
"""

factory.defaults['error.render_empty'] = False
factory.doc['props']['error.render_empty'] = """\
Render tag even if there is no error message.
"""

factory.defaults['error.message_tag'] = 'div'
factory.doc['props']['error.message_tag'] = """\
HTML tag to use to enclose each error message.
"""

factory.defaults['error.message_class'] = 'errormessage'
factory.doc['props']['error.message_class'] = """\
CSS class to apply to inner message-tag.
"""

factory.defaults['error.position'] = 'inner-before'
factory.doc['props']['error.position'] = """\
Error can be rendered at 3 different positions: ``before``/ ``after`` the
prior rendered output or with ``inner-before``/ ``inner-after``  it puts the
prior rendered output inside the tag used for the error message (beofre or
after the message.
"""



from node.utils import UNSET
from yafowil.base import factory
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase


class TestText(YafowilTestCase):

    def test_text_blueprint(self):
        # Regular text input
        widget = factory(
            'text',
            name='MYTEXT',
            value='Test Text "Some Text"')
        self.assertEqual(widget(), (
            '<input class="text" id="input-MYTEXT" name="MYTEXT" type="text" '
            'value="Test Text &quot;Some Text&quot;" />'
        ))

        widget.mode = 'display'
        self.assertEqual(widget(), (
            '<div class="display-text" id="display-MYTEXT">'
            'Test Text "Some Text"</div>'
        ))

        # Render with title attribute
        widget = factory(
            'text',
            name='MYTEXT',
            value='ja ha!',
            props={
                'title': 'My awesome title'
            })
        self.assertEqual(widget(), (
            '<input class="text" id="input-MYTEXT" name="MYTEXT" '
            'title="My awesome title" type="text" value="ja ha!" />'
        ))

        # Generic HTML5 Data
        widget = factory(
            'text',
            name='MYTEXT',
            value='ja ha!',
            props={
                'title': 'My awesome title',
                'data': {'foo': 'bar'}
            })
        self.assertEqual(widget(), (
            '<input class="text" data-foo=\'bar\' id="input-MYTEXT" '
            'name="MYTEXT" title="My awesome title" type="text" '
            'value="ja ha!" />'
        ))

        # Extract and persist
        widget = factory(
            'text',
            name='MYTEXT',
            props={
                'persist_writer': write_mapping_writer
            })
        data = widget.extract(request={'MYTEXT': '10'})
        self.assertEqual(data.name, 'MYTEXT')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, '10')

        model = dict()
        data.write(model)
        self.assertEqual(model, {'MYTEXT': '10'})



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.common import display_proxy_renderer
from yafowil.common import generic_display_renderer
from yafowil.common import generic_extractor
from yafowil.common import generic_required_extractor
from yafowil.common import input_generic_renderer
from yafowil.datatypes import generic_datatype_extractor
from yafowil.datatypes import generic_emptyvalue_extractor
from yafowil.utils import css_managed_props
from yafowil.utils import managedprops


###############################################################################
# text
###############################################################################

@managedprops(
    'data',
    'title',
    'size',
    'disabled',
    'autofocus',
    'placeholder',
    'autocomplete',
    *css_managed_props)
def text_edit_renderer(widget, data):
    return input_generic_renderer(widget, data)


factory.register(
    'text',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[text_edit_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ]
)

factory.doc['blueprint']['text'] = """\
One line text input blueprint.
"""

factory.defaults['text.type'] = 'text'
factory.doc['props']['text.type'] = """\
Type of input tag.
"""

factory.defaults['text.required_class'] = 'required'

factory.defaults['text.default'] = ''

factory.defaults['text.class'] = 'text'

factory.defaults['text.disabled'] = False
factory.doc['props']['text.disabled'] = """\
Flag  input field is disabled.
"""

factory.defaults['text.persist'] = True



from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.tests import wrapped_fxml


class TestCommon(YafowilTestCase):

    def test_emptyvalue_extraction(self):
        # Empty values
        widget = factory(
            'text',
            name='MYTEXT',
            props={
                'title': 'Default tests',
                'data': {'foo': 'bar'},
                'default': 'defaultvalue'
            })
        self.assertEqual(widget(), (
            '<input class="text" data-foo=\'bar\' id="input-MYTEXT" '
            'name="MYTEXT" title="Default tests" type="text" '
            'value="defaultvalue" />'
        ))

        data = widget.extract(request={})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract(request={'MYTEXT': ''})
        self.assertEqual(data.extracted, '')

        widget.attrs['emptyvalue'] = 'emptyvalue'
        data = widget.extract(request={'MYTEXT': ''})
        self.assertEqual(data.extracted, 'emptyvalue')

        widget.attrs['emptyvalue'] = False
        data = widget.extract(request={})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract(request={'MYTEXT': ''})
        self.assertFalse(data.extracted, False)

        widget.attrs['emptyvalue'] = UNSET
        data = widget.extract(request={})
        self.assertEqual(data.extracted, UNSET)

        data = widget.extract(request={'MYTEXT': ''})
        self.assertEqual(data.extracted, UNSET)

    def test_autofocus_property(self):
        # Widget with autofocus property
        widget = factory(
            'text',
            name='AUTOFOCUS',
            value='',
            props={
                'autofocus': True
            })
        self.assertEqual(widget(), (
            '<input autofocus="autofocus" class="text" id="input-AUTOFOCUS" '
            'name="AUTOFOCUS" type="text" value="" />'
        ))

    def test_placeholder_property(self):
        # Widget with placeholder property
        widget = factory(
            'text',
            name='PLACEHOLDER',
            value='',
            props={
                'placeholder': 'This is a placeholder.'
            })
        self.assertEqual(widget(), (
            '<input class="text" id="input-PLACEHOLDER" name="PLACEHOLDER" '
            'placeholder="This is a placeholder." type="text" value="" />'
        ))

    def test_required_extractor(self):
        # Widget with requires input
        widget = factory(
            'text',
            name='REQUIRED',
            value='',
            props={
                'required': True,
                'error_class': True
            })
        self.assertEqual(widget(), (
            '<input class="required text" id="input-REQUIRED" name="REQUIRED" '
            'required="required" type="text" value="" />'
        ))

        # Extract with empty request, key not in request therefore no error
        data = widget.extract({})
        self.assertEqual(data.name, 'REQUIRED')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])

        # Extract with empty input sent, required error expected
        data = widget.extract({'REQUIRED': ''})
        self.assertEqual(data.name, 'REQUIRED')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('Mandatory field was empty')]
        )

        # With getter value set, empty request, no error expected
        widget = factory(
            'text',
            name='REQUIRED',
            value='Test Text',
            props={
                'required': True,
                'error_class': True
            })
        data = widget.extract({})
        self.assertEqual(data.name, 'REQUIRED')
        self.assertEqual(data.value, 'Test Text')
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])
        self.assertEqual(widget(data=data), (
            '<input class="required text" id="input-REQUIRED" name="REQUIRED" '
            'required="required" type="text" value="Test Text" />'
        ))

        # With getter value set, request given, error expected
        data = widget.extract({'REQUIRED': ''})
        self.assertEqual(data.name, 'REQUIRED')
        self.assertEqual(data.value, 'Test Text')
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('Mandatory field was empty')]
        )
        self.assertEqual(widget(data=data), (
            '<input class="error required text" id="input-REQUIRED" '
            'name="REQUIRED" required="required" type="text" value="" />'
        ))

        # Create a custom error message
        widget = factory(
            'text',
            name='REQUIRED',
            value='',
            props={
                'required': 'You fool, fill in a value!'
            })
        data = widget.extract({'REQUIRED': ''})
        self.assertEqual(data.name, 'REQUIRED')
        self.assertEqual(data.value, '')
        self.assertEqual(data.extracted, '')
        self.assertEqual(
            data.errors,
            [ExtractionError('You fool, fill in a value!')]
        )

        # ``required`` property could be a callable as well
        def required_callback(widget, data):
            return u"Foooo"

        widget = factory(
            'text',
            name='REQUIRED',
            value='',
            props={
                'required': required_callback
            })
        data = widget.extract({'REQUIRED': ''})
        self.assertEqual(data.errors, [ExtractionError('Foooo')])

    def test_generic_display_renderer(self):
        # Display mode of text widget uses ``generic_display_renderer``
        widget = factory(
            'text',
            name='DISPLAY',
            value='lorem ipsum',
            mode='display')
        self.assertEqual(
            widget(),
            '<div class="display-text" id="display-DISPLAY">lorem ipsum</div>'
        )

        widget = factory(
            'text',
            name='DISPLAY',
            value=123.4567890,
            mode='display',
            props={
                'template': '%0.3f'
            })
        self.assertEqual(
            widget(),
            '<div class="display-text" id="display-DISPLAY">123.457</div>'
        )

        def mytemplate(widget, data):
            return '<TEMPLATE>%s</TEMPLATE>' % data.value

        widget = factory(
            'text',
            name='DISPLAY',
            value='lorem ipsum',
            mode='display',
            props={
                'template': mytemplate
            })
        self.checkOutput("""
        <div class="display-text" id="display-DISPLAY">
          <TEMPLATE>lorem ipsum</TEMPLATE>
        </div>
        """, fxml(widget()))

        # ``display_proxy`` can be used if mode is 'display' to proxy the value
        # in a hidden field
        widget = factory(
            'text',
            name='DISPLAY',
            value='lorem ipsum',
            mode='display',
            props={
                'display_proxy': True
            })
        self.checkOutput("""
        <div>
          <div class="display-text" id="display-DISPLAY">lorem ipsum</div>
          <input class="text" id="input-DISPLAY" name="DISPLAY" type="hidden"
                 value="lorem ipsum"/>
        </div>
        """, wrapped_fxml(widget()))

        # On widgets with display mode display_proxy property set, the data
        # gets extracted
        data = widget.extract(request={'DISPLAY': 'lorem ipsum'})
        self.assertEqual(data.name, 'DISPLAY')
        self.assertEqual(data.value, 'lorem ipsum')
        self.assertEqual(data.extracted, 'lorem ipsum')

        # Skip mode renders empty string
        widget = factory(
            'text',
            name='SKIP',
            value='lorem ipsum',
            mode='skip')
        self.assertEqual(widget(), '')

    def test_bytes_datatype_rendering_and_extraction(self):
        widget = factory(
            'text',
            value=b'\r\n\x01\x9a\x03\xff',
            name='BYTES',
            props={
                'datatype': bytes
            })
        self.assertEqual(widget(), (
            u'<input class="text" id="input-BYTES" name="BYTES" '
            u'type="text" value="\\r\\n\\x01\\x9a\\x03\\xff" />'
        ))

        widget = factory(
            'text',
            name='BYTES',
            props={
                'datatype': bytes
            })
        data = widget.extract({'BYTES': u'\\r\\n\\x01\\x9a\\x03\\xff'})
        self.assertEqual(data.extracted, b'\r\n\x01\x9a\x03\xff')

    def test_BC_imports(self):
        # button
        from yafowil.common import submit_renderer
        from yafowil.common import button_renderer

        # checkbox
        from yafowil.common import checkbox_extractor
        from yafowil.common import checkbox_edit_renderer
        from yafowil.common import checkbox_display_renderer

        # datatypes
        from yafowil.common import generic_emptyvalue_extractor
        from yafowil.common import generic_datatype_extractor
        from yafowil.common import DATATYPE_LABELS

        # email
        from yafowil.common import email_extractor

        # field
        from yafowil.common import field_renderer
        from yafowil.common import label_renderer
        from yafowil.common import help_renderer
        from yafowil.common import error_renderer

        # file
        from yafowil.common import file_extractor
        from yafowil.common import mimetype_extractor
        from yafowil.common import input_file_edit_renderer
        from yafowil.common import convert_bytes
        from yafowil.common import input_file_display_renderer
        from yafowil.common import file_options_renderer

        # lines
        from yafowil.common import lines_extractor
        from yafowil.common import lines_edit_renderer
        from yafowil.common import lines_display_renderer

        # number
        from yafowil.common import number_extractor

        # password
        from yafowil.common import minlength_extractor
        from yafowil.common import ascii_extractor
        from yafowil.common import password_extractor
        from yafowil.common import password_edit_renderer
        from yafowil.common import password_display_renderer

        # proxy
        from yafowil.common import input_proxy_renderer

        # select
        from yafowil.common import select_extractor
        from yafowil.common import select_exists_marker
        from yafowil.common import select_edit_renderer_props
        from yafowil.common import select_block_edit_renderer
        from yafowil.common import select_cb_edit_renderer
        from yafowil.common import select_edit_renderer
        from yafowil.common import select_display_renderer

        # tag
        from yafowil.common import tag_renderer

        # text
        from yafowil.common import text_edit_renderer

        # textarea
        from yafowil.common import textarea_managed_props
        from yafowil.common import textarea_attributes
        from yafowil.common import textarea_renderer

        # url
        from yafowil.common import url_extractor

        # utils
        from yafowil.utils import convert_value_to_datatype
        from yafowil.utils import convert_values_to_datatype



# -*- coding: utf-8 -*-
import yafowil.utils


@yafowil.utils.entry_point(order=0)
def register():  # noqa
    import yafowil.button
    import yafowil.checkbox
    import yafowil.common
    import yafowil.compound
    import yafowil.datatypes
    import yafowil.email
    import yafowil.field
    import yafowil.file
    import yafowil.hidden
    import yafowil.lines
    import yafowil.number
    import yafowil.password
    import yafowil.persistence
    import yafowil.proxy
    import yafowil.search
    import yafowil.select
    import yafowil.table
    import yafowil.tag
    import yafowil.text
    import yafowil.textarea
    import yafowil.url


# execute all register entry points. supposed to be used for widget and theme
# registration
for ep, cb in yafowil.utils.get_plugins('register'):
    cb()


# execute all configure entry points. supposed to be used for theme
# configuration, like setting factory defaults and defining macros.
for ep, cb in yafowil.utils.get_plugins('configure'):
    cb()  # pragma: no cover



# -*- coding: utf-8 -*-
from node.base import AttributedNode
from node.base import OrderedNode
from node.behaviors import Attributes
from node.utils import UNSET
from plumber import plumbing
from yafowil.base import factory
from yafowil.tests import YafowilTestCase
from yafowil.utils import Tag
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import callable_value
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import data_attrs_helper
from yafowil.utils import generic_html5_attrs
from yafowil.utils import get_example
from yafowil.utils import get_example_names
from yafowil.utils import get_plugin_names
from yafowil.utils import get_plugins
from yafowil.utils import managedprops
from yafowil.utils import tag as deprecated_tag
from yafowil.utils import vocabulary


class TestUtils(YafowilTestCase):

    def test_entry_point(self):
        # Test entry_point support tools
        self.checkOutput("""
        [...(EntryPoint.parse('register = yafowil.loader:register'),
        <function register at ...)...]
        """, str(list(get_plugins())))

        self.assertEqual(list(get_plugins('nonexisting')), [])

        self.checkOutput("""
        [...'yafowil'...]
        """, str((get_plugin_names())))

        self.assertEqual(get_plugin_names('nonexisting'), [])

    def test_examples_lookup(self):
        # Test examples lookup
        self.checkOutput("""
        ['yafowil'...]
        """, str(sorted(get_example_names())))

        factory.register_macro('field', 'field:label:error', {})

        self.assertEqual(get_example('inexistent'), None)

        examples = get_example('yafowil')
        self.checkOutput("""
        Plain Text
        ----------
        ...
        """, examples[0]['doc'])

        self.assertEqual(examples[0]['title'], 'Plain Text')
        self.assertEqual(examples[0]['widget'].name, 'yafowil-plaintext')

    def test_vocabulary(self):
        # Test the Vocabulary
        self.assertEqual(vocabulary('foo'), [('foo', 'foo')])
        self.assertEqual(vocabulary({'key': 'value'}), [('key', 'value')])
        self.assertEqual(
            vocabulary(['value', 'value2']),
            [('value', 'value'), ('value2', 'value2')]
        )
        self.assertEqual(
            vocabulary([
                ('key', 'value'),
                ('key2', 'value2', 'v2.3'),
                ('key3',)
            ]),
            [('key', 'value'), ('key2', 'value2'), ('key3', 'key3')]
        )

        def callme():
            return 'bar'
        self.assertEqual(vocabulary(callme), [('bar', 'bar')])
        self.assertTrue(vocabulary(None) is None)

    def test_Tag(self):
        # Test Tag
        tag = Tag(lambda msg: msg)
        t = tag('p', b'Lorem Ipsum. ', u'Hello World!',
                class_=b'fancy', id=u'2f5b8a234ff')
        self.assertEqual(t, (
            u'<p class="fancy" id="2f5b8a234ff">Lorem Ipsum. Hello World!</p>'
        ))
        self.assertEqual(
            tag('dummy', name='foo', **{'data-foo': 'bar'}),
            u'<dummy data-foo=\'bar\' name="foo" />'
        )
        self.assertEqual(tag('dummy', name=None), u'<dummy />')
        self.assertEqual(tag('dummy', name=UNSET), u'<dummy />')
        # deprecated test
        self.assertEqual(deprecated_tag('div', 'foo'), u'<div>foo</div>')

    def test_cssid(self):
        # Test CSS id
        @plumbing(Attributes)
        class CSSTestNode(OrderedNode):
            @property
            def dottedpath(self):
                return u'.'.join([it for it in self.path if it])

        widget = CSSTestNode(name='form')

        widget.attrs['structural'] = True
        self.assertEqual(cssid(widget, 'PREFIX'), None)

        child = widget['child'] = CSSTestNode()
        self.assertEqual(cssid(child, 'PREFIX'), 'PREFIX-form-child')
        self.assertEqual(
            cssid(child, 'PREFIX', postfix='POSTFIX'),
            'PREFIX-form-child-POSTFIX'
        )

        child = widget[u'Hällo Wörld'] = CSSTestNode()
        self.assertEqual(cssid(child, 'PREFIX'), 'PREFIX-form-Hallo_World')

    def test_css_classes(self):
        # Test CSS Classes
        @plumbing(Attributes)
        class CSSTestNode(OrderedNode):
            pass

        widget = CSSTestNode()
        widget.attrs['required'] = False
        widget.attrs['required_class'] = None
        widget.attrs['required_class_default'] = 'required'
        widget.attrs['error_class'] = None
        widget.attrs['error_class_default'] = 'error'
        widget.attrs['valid_class'] = None
        widget.attrs['valid_class_default'] = 'valid'
        widget.attrs['class'] = None
        widget.attrs['class_add'] = None

        class DummyData(object):
            def __init__(self):
                self.errors = []
                self.extracted = UNSET

            @property
            def root(self):
                return self

        data = DummyData()

        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(cssclasses(widget, data), 'bar foo')

        widget.attrs['class'] = None
        widget.attrs['required'] = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.required = False
        data.errors = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['error_class'] = True
        self.assertEqual(cssclasses(widget, data), 'error')

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(cssclasses(widget, data), 'bar error foo')

        widget.attrs['class'] = lambda w, d: 'baz'
        self.assertEqual(cssclasses(widget, data), 'baz error')

        widget.attrs['class_add'] = lambda w, d: 'addclass_from_callable'
        self.assertEqual(
            cssclasses(widget, data),
            'addclass_from_callable baz error'
        )

        widget.attrs['class_add'] = 'addclass'
        self.assertEqual(cssclasses(widget, data), 'addclass baz error')

        widget.attrs['class'] = None
        widget.attrs['class_add'] = None
        widget.attrs['error_class'] = 'othererror'
        self.assertEqual(cssclasses(widget, data), 'othererror')

        data.errors = False
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['required'] = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['required_class'] = True
        self.assertEqual(cssclasses(widget, data), 'required')

        widget.attrs['required_class'] = 'otherrequired'
        self.assertEqual(cssclasses(widget, data), 'otherrequired')

        widget.attrs['error_class'] = True
        data.errors = True
        widget.attrs['required_class'] = 'required'
        self.assertEqual(cssclasses(widget, data), 'error required')

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(
            cssclasses(widget, data),
            'bar error foo required'
        )

        self.assertEqual(
            cssclasses(widget, data, additional=['zika', 'akiz']),
            'akiz bar error foo required zika'
        )

        data.errors = []
        data.extracted = True
        widget.attrs['required'] = False
        widget.attrs['class'] = None

        widget.attrs['valid_class'] = True
        self.assertEqual(cssclasses(widget, data), 'valid')

        widget.attrs['valid_class'] = 'custom_valid'
        self.assertEqual(cssclasses(widget, data), 'custom_valid')

    def test_managedprops(self):
        # Test managedprops annotation
        @managedprops('foo', 'bar')
        def somefunc(a, b, c):
            return a, b, c

        self.assertEqual(somefunc(1, 2, 3), (1, 2, 3))

        self.assertEqual(
            somefunc.__yafowil_managed_props__,
            ('foo', 'bar')
        )

    def test_attr_value(self):
        # Test attr_value
        widget = AttributedNode()
        data = AttributedNode()

        widget.attrs['attr'] = 'value'
        self.assertEqual(attr_value('attr', widget, data), 'value')

        def func_callback(widget, data):
            return 'func_callback value'

        widget.attrs['attr'] = func_callback
        self.assertEqual(
            attr_value('attr', widget, data),
            'func_callback value'
        )

        def failing_func_callback(widget, data):
            raise Exception('failing_func_callback')

        widget.attrs['attr'] = failing_func_callback
        with self.assertRaises(Exception) as arc:
            attr_value('attr', widget, data)
        self.assertEqual(str(arc.exception), 'failing_func_callback')

        class FormContext(object):
            def instance_callback(self, widget, data):
                return 'instance_callback'

            def failing_instance_callback(self, widget, data):
                raise Exception('failing_instance_callback')

        context = FormContext()
        widget.attrs['attr'] = context.instance_callback
        self.assertEqual(
            attr_value('attr', widget, data),
            'instance_callback'
        )

        widget.attrs['attr'] = context.failing_instance_callback
        with self.assertRaises(Exception) as arc:
            attr_value('attr', widget, data)
        self.assertEqual(str(arc.exception), 'failing_instance_callback')

    def test_as_data_attrs(self):
        self.assertTrue(as_data_attrs is generic_html5_attrs)
        html5_attrs = as_data_attrs({
            'foo': 'bar',
            'baz': ['bam'],
            'nada': None,
            'unset': UNSET
        })
        self.assertEqual(
            html5_attrs,
            {'data-baz': '["bam"]', 'data-foo': 'bar'}
        )
        html5_attrs = as_data_attrs(None)
        self.assertEqual(html5_attrs, {})
        html5_attrs = as_data_attrs(UNSET)
        self.assertEqual(html5_attrs, {})

    def test_data_attrs_helper(self):
        data = AttributedNode()
        widget = AttributedNode()
        widget.attrs['testattr1'] = 'value'
        widget.attrs['testattr2'] = True
        widget.attrs['testattr3'] = False
        widget.attrs['testattr4'] = None
        widget.attrs['testattr5'] = ['item1', 'item2', 'item3']
        widget.attrs['testattr6'] = {
            'key1': 'item1',
            'key2': 'item2',
            'key3': 'item3'
        }
        widget.attrs['testattr7'] = 1234
        widget.attrs['testattr8'] = 1234.5678
        widget.attrs['testattr9'] = UNSET
        widget.attrs['camelAttrName'] = 'camelValue'
        data_attrs_keys = [
            'testattr1', 'testattr2', 'testattr3', 'testattr4', 'testattr5',
            'testattr6', 'testattr7', 'testattr8', 'camelAttrName'
        ]
        data_attrs = data_attrs_helper(widget, data, data_attrs_keys)

        self.assertEqual(data_attrs['data-testattr1'], 'value')
        self.assertEqual(data_attrs['data-testattr2'], 'true')
        self.assertEqual(data_attrs['data-testattr3'], 'false')
        self.assertFalse('data-testattr4' in data_attrs)
        self.assertEqual(
            data_attrs['data-testattr5'],
            '["item1", "item2", "item3"]'
        )
        self.assertTrue(data_attrs['data-testattr6'].find('"key1": "item1"'))
        self.assertTrue(data_attrs['data-testattr6'].find('"key2": "item2"'))
        self.assertTrue(data_attrs['data-testattr6'].find('"key3": "item3"'))
        self.assertTrue(data_attrs['data-testattr6'].startswith('{'))
        self.assertTrue(data_attrs['data-testattr6'].endswith('}'))
        self.assertEqual(data_attrs['data-testattr7'], '1234')
        self.assertEqual(data_attrs['data-testattr8'], '1234.5678')
        self.assertFalse('data-testattr9' in data_attrs)
        self.assertEqual(data_attrs['data-camel-attr-name'], 'camelValue')

        # Test with Tag renderer
        tag = Tag(lambda msg: msg)
        self.checkOutput("""
        <dummy
          data-camel-attr-name='camelValue'
          data-testattr1='value'
          data-testattr2='true'
          data-testattr3='false'
          data-testattr5='["item1", "item2", "item3"]'
          data-testattr6='{"..."}'
          data-testattr7='1234'
          data-testattr8='1234.5678'
          name="foo" />
        """, tag('dummy', name='foo', **data_attrs))

    def test_callable_value(self):
        # non callbale is returned as is
        self.assertEqual(callable_value('1', None, None), '1')

        # callable expects widget and data
        def dummy_callable(widget, data):
            return "2"
        self.assertEqual(callable_value(dummy_callable, None, None), '2')

        # callable with no parameters raises type error
        def invalid_signature():
            pass  # pragma: no cover
        self.assertRaises(TypeError, invalid_signature, None, None)
